"""This module defines ORM classes for logs"""

from datetime import datetime
import enum
import uuid

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class LogLevel(enum.Enum):
    """An enum representing the level of logs."""
    TRACE = "Trace"
    INFO = "Info"
    ERROR = "Error"


class Log(Base):
    """A class representing log objects in the ORM."""
    __tablename__ = "Logs"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    log_time: Mapped[datetime] = mapped_column(default=datetime.now)
    log_level: Mapped[LogLevel]
    process_name: Mapped[str] = mapped_column(String(100))
    job_id: Mapped[uuid.UUID | None] = mapped_column(nullable=True)
    log_message: Mapped[str] = mapped_column(String(8000))

    def to_row_dict(self) -> dict[str, str]:
        """Convert log to a row dictionary for display in a table."""
        return {
            "Log Time": datetime_util.format_datetime(self.log_time),
            "Level": self.log_level.value,
            "Process Name": self.process_name,
            "Message": self.log_message,
            "Short Job ID": str(self.job_id)[:8],
            "Full Job ID": self.job_id,
            "ID": str(self.id)
        }



"""Class for queue element popups."""
import json
from typing import Callable

from nicegui import ui

from OpenOrchestrator.orchestrator import test_helper
from OpenOrchestrator.database import db_util
from OpenOrchestrator.orchestrator.popups import generic_popups
from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput
from OpenOrchestrator.database.queues import QueueStatus, QueueElement


# pylint: disable-next=too-few-public-methods, too-many-instance-attributes
class QueueElementPopup():
    """A popup to display queue element data.
    """
    def __init__(self, queue_element: QueueElement | None, on_dialog_close_callback: Callable, queue_name: str):
        """Show a dialogue with details of the row selected.

        Args:
            row_data: Data from the row selected.
        """
        self.on_dialog_close_callback = on_dialog_close_callback
        self.queue_element = queue_element
        self.queue_name = queue_name
        with ui.dialog() as self.dialog:
            with ui.card().style('min-width:  37.5rem; max-width: 50rem').classes('gap-0'):

                with ui.row().classes('w-full justify-between items-start'):
                    with ui.column().classes('gap-0'):
                        ui.label("ID:").classes('text-subtitle2 text-grey-7')
                        self.id_text = ui.label()
                        self.reference = ui.input("Reference")
                    with ui.column().classes('gap-0 items-end'):
                        self.created_by_label = ui.label("Created by:").classes('text-subtitle2 text-grey-7')
                        self.created_by = ui.label()
                        self.status = ui.select(options={status.name: status.value for status in QueueStatus}, label="Status").classes("w-32")

                with ui.column().classes('gap-0'):
                    with ui.row().classes('w-full'):
                        self.data_field = ui.textarea("Data").classes('w-full')
                    with ui.row().classes('w-full mt-4'):
                        self.message = ui.input('Message').classes('w-full')
                    with ui.row().classes('w-full mt-4'):
                        with ui.column().classes('flex-1'):
                            self.created_date = DatetimeInput("Created Date", allow_empty=True)
                        with ui.column().classes('flex-1'):
                            self.start_date = DatetimeInput("Start Date", allow_empty=True)
                        with ui.column().classes('flex-1'):
                            self.end_date = DatetimeInput("End Date", allow_empty=True)

                with ui.row().classes('w-full mt-4'):
                    self.save_button = ui.button(text='Save', on_click=self._save_and_close).classes('mt-4')
                    self.close_button = ui.button('Close', on_click=self._close_dialog).classes('mt-4')
                    self.delete_button = ui.button(text='Delete', on_click=self._delete_element, color="negative").classes('mt-4')

        test_helper.set_automation_ids(self, "queue_element_popup")
        self.dialog.open()
        self._pre_populate()

    def _pre_populate(self):
        """Pre populate the inputs with an existing credential."""
        if self.queue_element:
            if not self.queue_element.created_by:
                self.created_by_label.visible = False
            else:
                self.created_by.text = self.queue_element.created_by
            self.id_text.text = self.queue_element.id
            self.reference.value = self.queue_element.reference
            self.status.value = self.queue_element.status.name
            self.message.value = self.queue_element.message
            self.data_field.value = self._prettify_json(self.queue_element.data)
            self.created_date.set_datetime(self.queue_element.created_date)
            self.start_date.set_datetime(self.queue_element.start_date)
            self.end_date.set_datetime(self.queue_element.end_date)
        else:
            self.id_text.text = "NOT SAVED"
            self.created_by.text = "Orchestrator UI"
            self.status.value = "NEW"
            self.delete_button.visible = False

    def _prettify_json(self, json_string: str) -> str:
        if not json_string:
            return None
        try:
            data = json.loads(json_string)
            return json.dumps(data, indent=2, ensure_ascii=False)
        except ValueError:
            return json_string

    def _close_dialog(self):
        self.on_dialog_close_callback()
        self.dialog.close()

    async def _delete_element(self):
        if not self.queue_element:
            return
        if await generic_popups.question_popup(f"Delete element '{self.queue_element.id}'?", "Delete", "Cancel", color1='negative'):
            db_util.delete_queue_element(self.queue_element.id)
            ui.notify("Queue element deleted", type='positive')
            self._close_dialog()

    def _save_element(self):
        if not self.queue_element:
            self.queue_element = db_util.create_queue_element(self.queue_name)
            self.id_text.text = self.queue_element.id
            ui.notify("New queue element created", type="positive")
        db_util.update_queue_element(self.queue_element.id,
                                     reference=self.reference.value,
                                     status=self.status.value,
                                     data=self.data_field.value,
                                     message=self.message.value,
                                     created_by=self.created_by.text,
                                     created_date=self.created_date.get_datetime(),
                                     start_date=self.start_date.get_datetime(),
                                     end_date=self.end_date.get_datetime())

    def _save_and_close(self):
        self._save_element()
        self._close_dialog()



"""Database revision '1c87ed320c78': Added jobs"""

from alembic import op
import sqlalchemy as sa

# pylint: disable=invalid-name
# revision identifiers, used by Alembic.
revision: str = '1c87ed320c78'
down_revision = '9698388a0709'
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade the database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'Jobs',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('process_name', sa.String(length=100), nullable=False),
        sa.Column('scheduler_name', sa.String(length=100), nullable=False),
        sa.Column('start_time', sa.DateTime(), nullable=False),
        sa.Column('end_time', sa.DateTime(), nullable=True),
        sa.Column('status', sa.Enum('RUNNING', 'DONE', 'FAILED', 'KILLED', name='jobstatus'), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.add_column('Logs', sa.Column('job_id', sa.Uuid(), nullable=True))



"""Module for truncating a string by removing the middle part."""

import math


def truncate_message(message: str, max_length: int = 8000) -> str:
    """Truncate a message to max_length characters

    Args:
        message: The string to truncate
        max_length: Maximum allowed length, defaults to 8000

    Returns:
        A string with length set to max_length
    """
    if len(message) <= max_length:
        return message
    return message[:math.ceil(max_length/2)] + message[-math.floor(max_length/2):]



"""This module contains tests of the functionality of alembic database migration."""

import unittest
import os
import subprocess
import sys

from OpenOrchestrator.database import db_util
from OpenOrchestrator.tests import db_test_util


class TestAlembic(unittest.TestCase):
    """Test functionality of Alembic database migrations."""
    def setUp(self) -> None:
        db_util.connect(os.environ["CONN_STRING"])
        db_test_util.drop_all_tables()

    def test_new_database(self):
        """Test creating a new database with Alembic"""
        subprocess.run([sys.executable, "-m", "OpenOrchestrator", "u", os.environ["CONN_STRING"], "--new"], check=True)


if __name__ == '__main__':
    unittest.main()



"""Database revision 'b67b7649b282': Initial"""

from alembic import op
import sqlalchemy as sa


# pylint: disable=invalid-name
# revision identifiers, used by Alembic.
revision: str = 'b67b7649b282'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade the database."""
    op.create_table(
        'Constants',
        sa.Column('name', sa.String(length=100), nullable=False),
        sa.Column('value', sa.String(length=1000), nullable=False),
        sa.Column('changed_at', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('name')
    )
    op.create_table(
        'Credentials',
        sa.Column('name', sa.String(length=100), nullable=False),
        sa.Column('username', sa.String(length=250), nullable=False),
        sa.Column('password', sa.String(length=1000), nullable=False),
        sa.Column('changed_at', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('name')
    )
    op.create_table(
        'Logs',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('log_time', sa.DateTime(), nullable=False),
        sa.Column('log_level', sa.Enum('TRACE', 'INFO', 'ERROR', name='loglevel'), nullable=False),
        sa.Column('process_name', sa.String(length=100), nullable=False),
        sa.Column('log_message', sa.String(length=8000), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Queues',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('queue_name', sa.String(length=100), nullable=False),
        sa.Column('status', sa.Enum('NEW', 'IN_PROGRESS', 'DONE', 'FAILED', 'ABANDONED', name='queuestatus'), nullable=False),
        sa.Column('data', sa.String(length=2000), nullable=True),
        sa.Column('reference', sa.String(length=100), nullable=True),
        sa.Column('created_date', sa.DateTime(), nullable=False),
        sa.Column('start_date', sa.DateTime(), nullable=True),
        sa.Column('end_date', sa.DateTime(), nullable=True),
        sa.Column('message', sa.String(length=1000), nullable=True),
        sa.Column('created_by', sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_Queues_created_date'), 'Queues', ['created_date'], unique=False)
    op.create_index(op.f('ix_Queues_queue_name'), 'Queues', ['queue_name'], unique=False)
    op.create_table(
        'Triggers',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('trigger_name', sa.String(length=100), nullable=False),
        sa.Column('process_name', sa.String(length=100), nullable=False),
        sa.Column('last_run', sa.DateTime(), nullable=True),
        sa.Column('process_path', sa.String(length=250), nullable=False),
        sa.Column('process_args', sa.String(length=1000), nullable=True),
        sa.Column('process_status', sa.Enum('IDLE', 'RUNNING', 'FAILED', 'DONE', 'PAUSED', 'PAUSING', name='triggerstatus'), nullable=False),
        sa.Column('is_git_repo', sa.Boolean(), nullable=False),
        sa.Column('is_blocking', sa.Boolean(), nullable=False),
        sa.Column('type', sa.Enum('SINGLE', 'SCHEDULED', 'QUEUE', name='triggertype'), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Queue_Triggers',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('queue_name', sa.String(length=200), nullable=False),
        sa.Column('min_batch_size', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['id'], ['Triggers.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Scheduled_Triggers',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('cron_expr', sa.String(length=200), nullable=False),
        sa.Column('next_run', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['id'], ['Triggers.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Single_Triggers',
        sa.Column('id', sa.Uuid(), nullable=False),
        sa.Column('next_run', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['id'], ['Triggers.id'], ),
        sa.PrimaryKeyConstraint('id')
    )



"""Database revision '9698388a0709': Added git branch to triggers"""

from alembic import op
import sqlalchemy as sa


# pylint: disable=invalid-name
# revision identifiers, used by Alembic.
revision: str = '9698388a0709'
down_revision = '90d46abd44a3'
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade the database."""
    op.add_column('Triggers', sa.Column('git_branch', sa.String(length=100), nullable=True))



"""This module defines ORM classes for queue elements."""

from datetime import datetime
import enum
from typing import Optional
import uuid

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class QueueStatus(enum.Enum):
    """An enum representing the status of a queue element."""
    NEW = 'New'
    IN_PROGRESS = 'In Progress'
    DONE = 'Done'
    FAILED = 'Failed'
    ABANDONED = 'Abandoned'


class QueueElement(Base):
    """A class representing a queue element in the ORM."""
    __tablename__ = "Queues"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    queue_name: Mapped[str] = mapped_column(String(100), index=True)
    status: Mapped[QueueStatus] = mapped_column(default=QueueStatus.NEW)
    data: Mapped[Optional[str]] = mapped_column(String(2000))
    reference: Mapped[Optional[str]] = mapped_column(String(100))
    created_date: Mapped[datetime] = mapped_column(default=datetime.now, index=True)
    start_date: Mapped[Optional[datetime]]
    end_date: Mapped[Optional[datetime]]
    message: Mapped[Optional[str]] = mapped_column(String(1000))
    created_by: Mapped[Optional[str]] = mapped_column(String(100))

    def to_row_dict(self) -> dict:
        """Convert the object to a dict for display in a table."""
        return {
            "ID": self.id,
            "Reference": self.reference,
            "Status": self.status,
            "Data": self.data,
            "Created Date": datetime_util.format_datetime(self.created_date),
            "Start Date": datetime_util.format_datetime(self.start_date),
            "End Date": datetime_util.format_datetime(self.end_date),
            "Message": self.message,
            "Created By": self.created_by
        }



"""This module is responsible for the layout and functionality of the Queues tab
in Orchestrator."""
from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.queues import QueueStatus
from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput
from OpenOrchestrator.orchestrator import test_helper
from OpenOrchestrator.orchestrator.popups.queue_element_popup import QueueElementPopup


QUEUE_COLUMNS = [
    {'name': "Queue Name", 'label': "Queue Name", 'field': "Queue Name", 'align': 'left', 'sortable': True},
    {'name': "New", 'label': "New", 'field': "New", 'align': 'left', 'sortable': True},
    {'name': "In Progress", 'label': "In Progress", 'field': "In Progress", 'align': 'left', 'sortable': True},
    {'name': "Done", 'label': "Done", 'field': "Done", 'align': 'left', 'sortable': True},
    {'name': "Failed", 'label': "Failed", 'field': "Failed", 'align': 'left', 'sortable': True},
    {'name': "Abandoned", 'label': "Abandoned", 'field': "Abandoned", 'align': 'left', 'sortable': True}
]

ELEMENT_COLUMNS = [
    {'name': "Reference", 'label': "Reference", 'field': "Reference", 'align': 'left', 'sortable': True},
    {'name': "Status", 'label': "Status", 'field': "Status", 'align': 'left', 'sortable': True},
    {'name': "Data", 'label': "Data", 'field': "Data", 'align': 'left', 'sortable': True, 'style': 'max-width: 200px; overflow: hidden; text-overflow: ellipsis;'},
    {'name': "Message", 'label': "Message", 'field': "Message", 'align': 'left', 'sortable': True},
    {'name': "Created Date", 'label': "Created Date", 'field': "Created Date", 'align': 'left', 'sortable': True},
    {'name': "Start Date", 'label': "Start Date", 'field': "Start Date", 'align': 'left', 'sortable': True},
    {'name': "End Date", 'label': "End Date", 'field': "End Date", 'align': 'left', 'sortable': True},
    {'name': "Created By", 'label': "Created By", 'field': "Created By", 'align': 'left', 'sortable': True},
    {'name': "ID", 'label': "ID", 'field': "ID", 'align': 'left', 'sortable': True}
]


# pylint: disable-next=too-few-public-methods
class QueueTab:
    """The 'Queues' tab object. It contains tables and buttons for dealing with queues."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            self.queue_table = ui.table(title="Queues", columns=QUEUE_COLUMNS, rows=[], row_key='Queue Name', pagination={'rowsPerPage': 50, 'sortBy': 'Queue Name'}).classes("w-full")
            self.queue_table.on("rowClick", self._row_click)
        test_helper.set_automation_ids(self, "queues_tab")

    def update(self):
        """Update the queue table with data from the database."""
        queue_count = db_util.get_queue_count()

        # Convert queue count to row elements
        rows = []
        for queue_name, count in queue_count.items():
            row = {
                "Queue Name": queue_name,
                "New": count.get(QueueStatus.NEW, 0),
                "In Progress": count.get(QueueStatus.IN_PROGRESS, 0),
                "Done": count.get(QueueStatus.DONE, 0),
                "Failed": count.get(QueueStatus.FAILED, 0),
                "Abandoned": count.get(QueueStatus.ABANDONED, 0),
            }
            rows.append(row)

        self.queue_table.update_rows(rows)

    def _row_click(self, event):
        row = event.args[1]
        queue_name = row["Queue Name"]
        QueuePopup(queue_name, self.update)


# pylint: disable-next=too-few-public-methods, too-many-instance-attributes
class QueuePopup:
    """A popup that displays queue elements in a queue."""
    def __init__(self, queue_name: str, update_callback):
        self.queue_name = queue_name
        self.order_by = "Created Date"
        self.order_descending = False
        self.page = 1
        self.rows_per_page = 25
        self.queue_count = 100
        self.update_callback = update_callback  # To make sure the main table updates changes to queue elements.

        with ui.dialog(value=True).props('full-width full-height') as dialog, ui.card():
            with ui.row().classes("w-full"):
                self.search_input = ui.input(label='Search', placeholder="Ref, message or data", on_change=self._update).style('margin-left: 1rem')
                self.status_select = ui.select(
                    options= {'All': 'All'} | {status.name: status.value for status in QueueStatus},
                    label="Status",
                    value="All",
                    on_change=self._update).classes("w-24")
                self.from_input = DatetimeInput("From Date", on_change=self._update, allow_empty=True)
                self.to_input = DatetimeInput("To Date", on_change=self._update, allow_empty=True)

                ui.space()

                ui.switch("Dense", on_change=lambda e: self._dense_table(e.value))
                self._create_column_filter()
                ui.button(icon='refresh', on_click=self._update)
                self.close_button = ui.button(icon="close", on_click=lambda: (dialog.close(), self.update_callback()))
            with ui.scroll_area().classes("h-full"):
                self.table = ui.table(columns=ELEMENT_COLUMNS, rows=[], row_key='ID', title=queue_name, pagination={'rowsPerPage': self.rows_per_page, 'rowsNumber': self.queue_count})
                self.table.classes("w-full sticky-header h-[calc(100vh-200px)] overflow-auto")
                self.table.on('rowClick', lambda e: self._open_queue_element_popup(e.args[1]))
                self.table.props(
                    ":rows-per-page-options='[10, 25, 50, 100, 1000]' rows-per-page-label='Queue elements per page:'")
                self.table.on('request', self._on_table_request)

                with self.table.add_slot("top"):
                    ui.label(self.queue_name).classes("text-xl")
                    ui.space()
                    self.new_button = ui.button(icon='playlist_add', on_click=self._open_create_dialog)

        self._update()
        test_helper.set_automation_ids(self, "queue_popup")

    def _dense_table(self, value: bool):
        """Change if the table is dense or not."""
        if value:
            self.table.props("dense")
        else:
            self.table.props(remove="dense")

    def _create_column_filter(self):
        """Create a menu with switches for toggling columns on and off."""
        def toggle(column: dict, visible: bool) -> None:
            column['classes'] = '' if visible else 'hidden'
            column['headerClasses'] = '' if visible else 'hidden'
            self.table.update()

        with ui.button("Columns", icon="menu"):
            with ui.menu(), ui.column().classes('gap-0 p-2'):
                for column in ELEMENT_COLUMNS:
                    ui.switch(column['label'], value=True, on_change=lambda e, column=column: toggle(column, e.value))

    def _update(self):
        """Update the table with values from the database."""
        search_input = self.search_input.value.strip()
        if len(search_input) == 0:
            search_input = None
        status = None if self.status_select.value == "All" else self.status_select.value.strip()

        from_date = self.from_input.get_datetime()
        to_date = self.to_input.get_datetime()
        offset = (self.page - 1) * self.rows_per_page
        order_by = str(self.order_by).lower().replace(" ", "_")

        queue_elements, queue_count = db_util.get_queue_elements(self.queue_name, status=status, limit=self.rows_per_page, from_date=from_date, to_date=to_date, order_by=order_by, order_desc=self.order_descending, offset=offset, search_term=search_input, include_count=True)
        self._update_pagination(queue_count)
        rows = [element.to_row_dict() for element in queue_elements]
        self.table.update_rows(rows)

    def _on_table_request(self, e):
        """Called when updating table pagination and sorting, to handle these manually and allow for server side pagination.

        Args:
            e: The event triggering the request.
        """
        pagination = e.args['pagination']
        self.page = pagination.get('page')
        self.rows_per_page = pagination.get('rowsPerPage')
        self.order_by = pagination.get('sortBy')
        self.order_descending = pagination.get('descending', False)
        self._update()

    def _update_pagination(self, queue_count):
        """Update pagination element.

        Args:
            queue_count: The element count of the current filtered table.
        """
        self.queue_count = queue_count
        self.table.pagination = {"rowsNumber": self.queue_count, "page": self.page, "rowsPerPage": self.rows_per_page, "sortBy": self.order_by, "descending": self.order_descending}

    def _open_queue_element_popup(self, row_data: ui.row):
        """Open editable popup for specified row.

        Args:
            row_data: Row data from row clicked.
        """
        queue_element = db_util.get_queue_element(row_data.get("ID"))
        QueueElementPopup(queue_element=queue_element, on_dialog_close_callback=self._update, queue_name=self.queue_name)

    def _open_create_dialog(self):
        QueueElementPopup(None, on_dialog_close_callback=self._update, queue_name=self.queue_name)



"""This module is responsible for the layout and functionality of the Trigger tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import SingleTrigger, ScheduledTrigger, QueueTrigger, TriggerType
from OpenOrchestrator.orchestrator.popups.trigger_popup import TriggerPopup
from OpenOrchestrator.orchestrator import test_helper

COLUMNS = [
    {'name': "Trigger Name", 'label': "Trigger Name", 'field': "Trigger Name", 'align': 'left', 'sortable': True},
    {'name': "Priority", 'label': "Priority", 'field': "Priority", 'align': 'left', 'sortable': True, ':sort': '(a, b, rowA, rowB) => b-a'},
    {'name': "Type", 'label': "Type", 'field': "Type", 'align': 'left', 'sortable': True},
    {'name': "Status", 'label': "Status", 'field': "Status", 'align': 'left', 'sortable': True},
    {'name': "Process Name", 'label': "Process Name", 'field': "Process Name", 'align': 'left', 'sortable': True},
    {'name': "Last Run", 'label': "Last Run", 'field': "Last Run", 'align': 'left', 'sortable': True},
    {'name': "Next_Run", 'label': "Next Run", 'field': "Next Run", 'align': 'left', 'sortable': True},
    {'name': "ID", 'label': "ID", 'field': "ID", 'align': 'left', 'sortable': True}
]


# pylint disable-next=too-few-public-methods
class TriggerTab():
    """The 'Trigger' tab object. It contains tables and buttons for dealing with triggers."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            with ui.row():
                self.single_button = ui.button("New Single Trigger", icon="add", on_click=lambda e: TriggerPopup(self, TriggerType.SINGLE))
                self.scheduled_button = ui.button("New Scheduled Trigger", icon="add", on_click=lambda e: TriggerPopup(self, TriggerType.SCHEDULED))
                self.queue_button = ui.button("New Queue Trigger", icon="add", on_click=lambda e: TriggerPopup(self, TriggerType.QUEUE))

            self.trigger_table = ui.table(columns=COLUMNS, rows=[], title="Triggers", pagination={'rowsPerPage': 50, 'sortBy': 'Trigger Name'}, row_key='ID').classes("w-full")
            self.trigger_table.on('rowClick', self._row_click)
            self.add_column_colors()

        test_helper.set_automation_ids(self, "trigger_tab")

    def _row_click(self, event):
        """Callback for when a row is clicked in the table."""
        row = event.args[1]
        trigger_id = row["ID"]
        trigger = db_util.get_trigger(trigger_id)

        if isinstance(trigger, SingleTrigger):
            TriggerPopup(self, TriggerType.SINGLE, trigger)
        elif isinstance(trigger, ScheduledTrigger):
            TriggerPopup(self, TriggerType.SCHEDULED, trigger)
        elif isinstance(trigger, QueueTrigger):
            TriggerPopup(self, TriggerType.QUEUE, trigger)

    def update(self):
        """Updates the tab and it's data."""
        triggers = db_util.get_all_triggers()
        self.trigger_table.rows = [t.to_row_dict() for t in triggers]
        self.trigger_table.update()

    def add_column_colors(self):
        """Add custom coloring to the trigger table."""
        # Add coloring to the status column
        color_dict = "{Running: 'green', Pausing: 'orange', Paused: 'orange', Failed: 'red', Killing: 'orange', Killed: 'grey-9'}"

        self.trigger_table.add_slot(
            "body-cell-Status",
            f'''
            <q-td key="Status" :props="props">
                <q-badge v-if="{color_dict}[props.value]" :color="{color_dict}[props.value]">
                    {{{{props.value}}}}
                </q-badge>
                <p v-else>
                    {{{{props.value}}}}
                </p>
            </q-td>
            '''
        )

        # Add coloring to 'Next run' column
        # If the next run is in the past the value should be red.
        # Use a very ugly parser to create a date from a dd-MM-yyyy HH:mm:ss date string.
        self.trigger_table.add_slot(
            "body-cell-Next_Run",
            '''
            <q-td key="Next_Run" :props="props">
                {{props.value}}
                <q-badge v-if="new Date(+props.value.substr(6,4), +props.value.substr(3,2)-1, +props.value.substr(0,2), +props.value.substr(11,2), +props.value.substr(14,2)) < new Date()" color='red'>
                    Overdue
                </q-badge>
            </q-td>
            '''
        )



"""This module contains tests for the trigger polling part of the Scheduler."""

import unittest
from unittest.mock import MagicMock, patch
from datetime import datetime

from OpenOrchestrator.scheduler import runner
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import SingleTrigger
from OpenOrchestrator.tests import db_test_util


@patch("OpenOrchestrator.scheduler.util.get_scheduler_name", return_value="Machine")
class TestTriggerPolling(unittest.TestCase):
    """Test the trigger polling functionality of the Scheduler."""
    def setUp(self) -> None:
        db_test_util.establish_clean_database()

    def test_poll_triggers_single(self, *_):
        """Test polling single triggers."""
        mock_app = setup_mock_app(has_running_jobs=False, is_exclusive=False)

        # Test with no triggers
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        # Test with idle trigger
        db_util.create_single_trigger("Future single trigger", "", datetime(2100, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        # Test with overdue trigger
        db_util.create_single_trigger("Past single trigger", "", datetime(2020, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertIsInstance(trigger, SingleTrigger)
        self.assertEqual(trigger.trigger_name, "Past single trigger")

    def test_poll_triggers_priority(self, *_):
        """Test polling triggers with priorities set."""
        mock_app = setup_mock_app(has_running_jobs=False, is_exclusive=False)

        db_util.create_single_trigger("Low", "", datetime(2000, 1, 1), "", "", False, False, 0)
        db_util.create_single_trigger("High", "", datetime(2000, 1, 1), "", "", False, False, 100)

        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "High")

        db_util.create_single_trigger("Higher Future", "", datetime(2100, 1, 1), "", "", False, False, 200)

        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "High")

    def test_poll_triggers_mixed(self, *_):
        """Test mixed trigger types with priorities set.
        Equal priorities should be prioritied by trigger type:
        Single > Scheduled > Queue
        """
        mock_app = setup_mock_app(has_running_jobs=False, is_exclusive=False)

        db_util.create_queue_trigger("Queue", "", "Queue", "", "", False, False, 1, 0)
        db_util.create_queue_element("Queue")
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Queue")

        db_util.create_scheduled_trigger("Scheduled", "", "", datetime(2000, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Scheduled")

        db_util.create_single_trigger("Single", "", datetime(2000, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Single")

        db_util.create_queue_trigger("Queue High", "", "Queue", "", "", False, False, 1, 1)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Queue High")

        db_util.create_scheduled_trigger("Scheduled High", "", "", datetime(2000, 1, 1), "", "", False, False, 1)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Scheduled High")

        db_util.create_single_trigger("Single High", "", datetime(2000, 1, 1), "", "", False, False, 1)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Single High")

    def test_poll_triggers_blocking(self, *_):
        """Test polling triggers when other jobs are running."""
        mock_app = setup_mock_app(has_running_jobs=True, is_exclusive=False)

        db_util.create_single_trigger("Single Blocking", "", datetime(2000, 1, 1), "", "", False, True, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        db_util.create_single_trigger("Single Non-blocking", "", datetime(2000, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Single Non-blocking")

    def test_poll_triggers_whitelist_non_exclusive(self, *_):
        """Test polling whitelisted triggers when Scheduler is not exclusive."""
        mock_app = setup_mock_app(has_running_jobs=False, is_exclusive=False)

        db_util.create_single_trigger("Single non machine", "", datetime(2000, 1, 1), "", "", False, False, 0, ["Non machine"])
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        db_util.create_single_trigger("Single no whitelist", "", datetime(2000, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Single no whitelist")

    def test_poll_triggers_whitelist_exclusive(self, *_):
        """Test polling whitelisted triggers when Scheduler is exclusive."""
        mock_app = setup_mock_app(has_running_jobs=False, is_exclusive=True)

        db_util.create_single_trigger("Single non machine", "", datetime(2000, 1, 1), "", "", False, False, 0, ["Non machine"])
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        db_util.create_single_trigger("Single no whitelist", "", datetime(2000, 1, 1), "", "", False, False, 0)
        trigger = runner.poll_triggers(mock_app)
        self.assertIsNone(trigger)

        db_util.create_single_trigger("Single whitelist", "", datetime(2000, 1, 1), "", "", False, False, 0, ["Machine"])
        trigger = runner.poll_triggers(mock_app)
        self.assertEqual(trigger.trigger_name, "Single whitelist")


def setup_mock_app(*, has_running_jobs: bool, is_exclusive: bool) -> MagicMock:
    """Create a mock Application object to be used in tests.

    Args:
        has_running_jobs: If the app should simulate having running jobs.
        is_exclusive: If the app should simulate only running whitelisted triggers.

    Returns:
        A mock Application object.
    """
    mock_app = MagicMock()
    mock_app.running_jobs = [None] if has_running_jobs else []
    mock_app.settings_tab_.whitelist_value.get.return_value = is_exclusive
    return mock_app


if __name__ == '__main__':
    unittest.main()



"""Tests relating to the queues tab in Orchestrator."""

import unittest
from datetime import datetime, timedelta

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.queues import QueueStatus
from OpenOrchestrator.tests.ui_tests import ui_util


class TestQueuesTab(unittest.TestCase):
    """Test functionality of the queues tab ui."""
    def setUp(self) -> None:
        self.browser = ui_util.open_orchestrator()
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queues_tab]").click()
        ui_util.refresh_ui(self.browser)

    def tearDown(self) -> None:
        self.browser.quit()

    @ui_util.screenshot_on_error
    def test_queues_table(self):
        """Test that queues are displayed correctly in the queues table."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "queues_tab_queue_table")
        self.assertEqual(table_data[0], ["Queue Name 1", "1", "2", "3", "4", "5"])
        self.assertEqual(table_data[1], ["Queue Name 2", "1", "2", "3", "4", "5"])

    @ui_util.screenshot_on_error
    def test_queue_popup(self):
        """Test that queue elements are displayed correctly in the queue popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)

        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")

        queue_elements = db_util.get_queue_elements("Queue Name 1")

        self.assertEqual(len(table_data), len(queue_elements))
        for i, queue_element in enumerate(queue_elements):
            self.assertIn(table_data[i][0], queue_element.reference)
            self.assertEqual(table_data[i][1], queue_element.status.value)
            self.assertEqual(table_data[i][2], queue_element.data)
            self.assertEqual(table_data[i][3], queue_element.message)
            self.assertRegex(table_data[i][4], r"(\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2})|(N/A)")
            self.assertRegex(table_data[i][5], r"(\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2})|(N/A)")
            self.assertRegex(table_data[i][6], r"(\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2})|(N/A)")
            self.assertEqual(table_data[i][7], queue_element.created_by)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_queue_popup_status_filter(self):
        """Test status filtering in queue popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        # Test each status individually
        for i, status in enumerate(QueueStatus):
            self._set_status_filter(status.value)
            table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
            expected_count = i + 1
            self.assertEqual(len(table_data), expected_count)

        # Test clearing status filter
        self._set_status_filter(None)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 15)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_queue_popup_search_filter(self):
        """Test search filtering in queue popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        # Test partial reference search
        self._set_search_filter("Reference 0")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        expected_count = len(QueueStatus)  # One "Reference 0,x" per status
        self.assertEqual(len(table_data), expected_count)
        self.assertIn("Reference 0", table_data[0][0])

        # Test exact reference search
        self._set_search_filter("Reference 0,1")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 1)
        self.assertEqual("Reference 0,1", table_data[0][0])

        # Test data search
        self._set_search_filter("Data 0,1")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 1)
        self.assertEqual("Data 0,1", table_data[0][2])

        # Test partial data search
        self._set_search_filter("ata 0")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 5)

        # Test message search
        self._set_search_filter("Message 0,1")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 1)
        self.assertEqual("Message 0,1", table_data[0][3])

        # Test no matches
        self._set_search_filter("NonExistentRef")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 0)

        # Test empty search (should show all)
        self._set_search_filter("")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 15)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_queue_popup_combined_filters(self):
        """Test combining multiple filters."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        yesterday = datetime.today() - timedelta(days=1)
        tomorrow = datetime.today() + timedelta(days=1)

        # Date + Status filter
        self._set_date_filter(from_date=yesterday, to_date=tomorrow)
        self._set_status_filter("New")  # Assuming this is first status
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 1)

        # Date + Search filter
        self._set_status_filter(None)  # Clear status
        self._set_search_filter("Reference 0")
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        expected_count = len(QueueStatus)
        self.assertIn("Reference 0", table_data[0][0])
        self.assertEqual(len(table_data), expected_count)

        # All three filters
        self._set_status_filter("In Progress")  # Assuming second status
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 1)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_queue_popup_filters(self):
        """Test setting filters in the queue popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)

        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        yesterday = datetime.today() - timedelta(days=1)
        tomorrow = datetime.today() + timedelta(days=1)

        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 15)

        # From yesterday
        self._set_date_filter(from_date=yesterday, to_date=None)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 15)

        # From tomorrow
        self._set_date_filter(from_date=tomorrow, to_date=None)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 0)

        # To yesterday
        self._set_date_filter(from_date=None, to_date=yesterday)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 0)

        # From yesterday to tomorrow
        self._set_date_filter(from_date=yesterday, to_date=tomorrow)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), 15)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_queue_element_popup(self):
        """Test content of queue element popup.
        """
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)

        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        ui_util.click_table_row(self.browser, "queue_popup_table", 0)
        for i, field in enumerate(['reference', 'status', 'data_field', 'message', 'created_date', 'start_date', 'end_date', 'created_by', 'id_text']):
            field_content = self.browser.find_element(By.CSS_SELECTOR, f"[auto-id=queue_element_popup_{field}]").get_attribute('value')
            if field in ['status', 'id_text', 'created_by']:
                field_content = self.browser.find_element(By.CSS_SELECTOR, f"[auto-id=queue_element_popup_{field}]").text
            if not field_content:
                field_content = "N/A"
            self.assertEqual(table_data[0][i], field_content)

    @ui_util.screenshot_on_error
    def test_create_new_queue_element(self):
        """Test creating a new queue element through the popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        # Click "New" button to create a new queue element
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_new_button]").click()

        # Fill in the form and save
        reference_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_reference]")
        reference_input.send_keys("New Reference")

        status_select = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_status]")
        status_select.click()
        status_select.find_element(By.XPATH, "//div[contains(@class,'q-item')]//span[text()='In Progress']").click()

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_save_button]").click()

        # Verify the new element appears in the queue popup table
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertTrue(any("New Reference" in row[0] for row in table_data))

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_edit_queue_element(self):
        """Test editing an existing queue element through the popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        # Open the first queue element in the popup
        ui_util.click_table_row(self.browser, "queue_popup_table", 0)

        # Edit the reference field
        reference_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_reference]")
        reference_input.clear()
        reference_input.send_keys("Edited Reference")

        # Change the status
        status_select = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_status]")
        status_select.click()
        status_select.find_element(By.XPATH, "//div[contains(@class,'q-item')]//span[text()='Done']").click()

        # Edit the date field
        reference_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_start_date]")
        reference_input.clear()
        reference_input.send_keys("01-01-2000 12:34:56")

        # Save the changes
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_save_button]").click()

        # Verify the changes in the queue popup table
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertTrue(any("Edited Reference" in row[0] for row in table_data))
        self.assertTrue(any("Done" in row[1] for row in table_data))
        self.assertTrue(any("01-01-2000 12:34:56" in row[5] for row in table_data))

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    @ui_util.screenshot_on_error
    def test_delete_queue_element(self):
        """Test deleting a queue element through the popup."""
        self._create_queue_elements()
        ui_util.refresh_ui(self.browser)
        ui_util.click_table_row(self.browser, "queues_tab_queue_table", 0)

        # Get the initial count of queue elements
        initial_count = len(ui_util.get_table_data(self.browser, "queue_popup_table"))

        # Open the first queue element in the popup
        ui_util.click_table_row(self.browser, "queue_popup_table", 0)

        # Click the delete button and confirm
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_element_popup_delete_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=popup_option1_button").click()

        # Verify the element was deleted
        table_data = ui_util.get_table_data(self.browser, "queue_popup_table")
        self.assertEqual(len(table_data), initial_count - 1)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_close_button]").click()

    def _create_queue_elements(self):
        """Create some queue elements.
        Creates 1x'New', 2x'In Progress' and so on.
        """
        for j in range(2):
            for i, status in enumerate(QueueStatus):
                for k in range(i+1):
                    qe = db_util.create_queue_element(f"Queue Name {j+1}", reference=f"Reference {k},{i}", data=f"Data {k},{i}", created_by=f"Creator {k},{i}")
                    db_util.set_queue_element_status(qe.id, status, message=f"Message {k},{i}")

    # This code is similar to some in test_logging_tab.
    # Since the ui might change it doesn't make sense to make a common function for this.
    # pylint: disable=duplicate-code
    def _set_date_filter(self, from_date: datetime | None, to_date: datetime | None):
        """Set the date filters."""
        # Clear filters
        from_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_from_input]")
        from_input.send_keys(Keys.CONTROL, "a", Keys.DELETE)
        to_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_to_input]")
        to_input.send_keys(Keys.CONTROL, "a", Keys.DELETE)

        if from_date:
            from_input.send_keys(from_date.strftime("%d-%m-%Y %H:%M:%S"))

        if to_date:
            to_input.send_keys(to_date.strftime("%d-%m-%Y %H:%M:%S"))

    def _set_status_filter(self, status=None):
        """Set status filter in queue popup."""
        if status is None:
            status = "All"
        status_select = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_status_select]")
        status_select.click()
        option = status_select.find_element(By.XPATH, f"//div[contains(@class,'q-item')]//span[text()='{status}']")
        option.click()

    def _set_search_filter(self, search_term=""):
        """Set reference search filter in queue popup."""
        search_field = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=queue_popup_search_input]")
        search_field.send_keys(Keys.CONTROL, "a", Keys.DELETE)
        if search_term:
            search_field.send_keys(search_term)


if __name__ == '__main__':
    unittest.main()



"""This module contains functions to help with ui tests."""

import subprocess
import os
import sys
import time
from typing import Callable
from pathlib import Path

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement

from OpenOrchestrator.orchestrator.application import get_free_port


ENCRYPTION_KEY = None


def open_orchestrator() -> webdriver.Chrome:
    """Open Orchestrator in a Selenium Chrome browser.

    Raises:
        RuntimeError: If the Orchestrator app didn't start.

    Returns:
        The Chrome browser logged in to Orchestrator.
    """
    conn_string = os.environ['CONN_STRING']
    os.environ['ORCHESTRATOR_TEST'] = "True"

    port = get_free_port()
    subprocess.Popen([sys.executable, "-m", "OpenOrchestrator", "o", "--port", str(port), "--dont_show"])  # pylint: disable=consider-using-with

    chrome_options = webdriver.ChromeOptions()
    custom_browser_options = [
        '--disable-search-engine-choice-screen',
        '--disable-dev-shm-usage',
        '--log-level=3',
        '--disable-logging'
        ]
    for option in custom_browser_options:
        chrome_options.add_argument(option)
    browser = webdriver.Chrome(options=chrome_options)
    browser.implicitly_wait(2)
    browser.maximize_window()
    for _ in range(5):
        try:
            browser.get(f"http://localhost:{port}")
            break
        except WebDriverException:
            pass
    else:
        raise RuntimeError("Couldn't connect to app after 5 tries.")

    # Wait for the ui to load
    time.sleep(1)

    conn_input = browser.find_element(By.CSS_SELECTOR, "input[auto-id=connection_frame_conn_input]")
    conn_input.send_keys(Keys.CONTROL, "a", Keys.DELETE)
    conn_input.send_keys(conn_string)

    time.sleep(1)

    browser.find_element(By.CSS_SELECTOR, "button[auto-id=settings_tab_key_button]").click()
    browser.find_element(By.CSS_SELECTOR, "button[auto-id=connection_frame_conn_button]").click()

    global ENCRYPTION_KEY  # pylint: disable=global-statement
    ENCRYPTION_KEY = browser.find_element(By.CSS_SELECTOR, "input[auto-id=connection_frame_key_input]").get_attribute("value")

    time.sleep(1)

    return browser


def refresh_ui(browser: webdriver.Chrome):
    """Press the refresh button."""
    browser.find_element(By.CSS_SELECTOR, "[auto-id=refresh_button").click()
    time.sleep(0.5)


def monkey_patch_send_keys():
    """Monkey patch the WebElement.send_keys function with a
    slower version. This works better with NiceGui elements.
    """
    old_send_keys = WebElement.send_keys

    def send_slow_keys(element: WebElement, *value: str | list):
        """Send keys to an element one at a time.
        If "value" is a list fall back to the default behavior.
        """
        if len(value) == 1 and isinstance(value[0], str):
            for c in value[0]:
                old_send_keys(element, c)
        else:
            old_send_keys(element, value)

    WebElement.send_keys = send_slow_keys


monkey_patch_send_keys()


def get_table_data(browser: webdriver.Chrome, auto_id: str) -> list[list[str]]:
    """Get the data from a table.

    Args:
        browser: The browser object to perform the action.
        auto_id: The automation id of the table.

    Returns:
        A 2D list of strings of all the data in the table.
    """
    table = browser.find_element(By.CSS_SELECTOR, f"[auto-id={auto_id}]")
    rows = table.find_element(By.TAG_NAME, "tbody").find_elements(By.TAG_NAME, "tr")

    data = []
    for row in rows:
        fields = row.find_elements(By.TAG_NAME, "td")
        data.append([f.text for f in fields])

    return data


def click_table_row(browser: webdriver.Chrome, auto_id: str, index: int):
    """Click a row in a table.

    Args:
        browser: The browser to perform the action.
        auto_id: The automation id of the table.
        index: The 0-based index of the row to click.
    """
    table = browser.find_element(By.CSS_SELECTOR, f"[auto-id={auto_id}]")
    rows = table.find_element(By.TAG_NAME, "tbody").find_elements(By.TAG_NAME, "tr")
    rows[index].click()
    time.sleep(0.5)


def screenshot_on_error(test_func: Callable) -> Callable:
    """A decorator that can be used on test functions to take a Selenium screenshot on errors.
    The unittest class must have a instance variable called "browser" which is the Selenium webdriver.

    Args:
        test_func: The test function to decorate.
    """
    def wrapper(*args, **kwargs):
        browser: webdriver.Chrome = args[0].browser

        try:
            test_func(*args, **kwargs)
        except Exception:
            folder = Path("error_screenshots")
            folder.mkdir(exist_ok=True)
            path = folder / Path(f"{test_func.__name__}.png")
            browser.save_screenshot(str(path))
            raise

    return wrapper



"""This module defines ORM classes for triggers."""

from datetime import datetime
import enum
from typing import Optional
import uuid

from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base
from OpenOrchestrator.database.data_types.string_list import StringList

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class TriggerStatus(enum.Enum):
    """An enum representing the status of trigger processes."""
    IDLE = "Idle"
    RUNNING = "Running"
    FAILED = "Failed"
    DONE = "Done"
    PAUSED = "Paused"
    PAUSING = "Pausing"
    KILLING = "Killing"
    KILLED = "Killed"


class TriggerType(enum.Enum):
    """An enum representing the type of triggers."""
    SINGLE = "Single"
    SCHEDULED = "Scheduled"
    QUEUE = "Queue"


class Trigger(Base):
    """A base class for all triggers in the ORM."""
    __tablename__ = "Triggers"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    trigger_name: Mapped[str] = mapped_column(String(100))
    process_name: Mapped[str] = mapped_column(String(100))
    last_run: Mapped[Optional[datetime]]
    process_path: Mapped[str] = mapped_column(String(250))
    process_args: Mapped[Optional[str]] = mapped_column(String(1000))
    process_status: Mapped[TriggerStatus] = mapped_column(default=TriggerStatus.IDLE)
    is_git_repo: Mapped[bool]
    git_branch: Mapped[Optional[str]] = mapped_column(String(100))
    is_blocking: Mapped[bool]
    scheduler_whitelist: Mapped[Optional[StringList]] = mapped_column(StringList(250))
    priority: Mapped[int] = mapped_column(default=0)
    type: Mapped[TriggerType]

    __mapper_args__ = {
        "polymorphic_on": "type",
        "polymorphic_abstract": True
    }

    def __repr__(self) -> str:
        return f"{self.trigger_name}: {self.type.value}"

    def to_row_dict(self) -> dict[str, str]:
        """Convert trigger to a row dictionary for display in a table."""
        return {
            "Trigger Name": self.trigger_name,
            "Type": self.type.value,
            "Status": self.process_status.value,
            "Process Name": self.process_name,
            "Last Run": datetime_util.format_datetime(self.last_run, "Never"),
            "ID": str(self.id),
            "Priority": str(self.priority)
        }


class SingleTrigger(Trigger):
    """A class representing single trigger objects in the ORM."""
    __tablename__ = "Single_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    next_run: Mapped[datetime]

    __mapper_args__ = {"polymorphic_identity": TriggerType.SINGLE}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = datetime_util.format_datetime(self.next_run)
        return row_dict


class ScheduledTrigger(Trigger):
    """A class representing scheduled trigger objects in the ORM."""
    __tablename__ = "Scheduled_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    cron_expr: Mapped[str] = mapped_column(String(200))
    next_run: Mapped[datetime]

    __mapper_args__ = {"polymorphic_identity": TriggerType.SCHEDULED}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = datetime_util.format_datetime(self.next_run)
        return row_dict


class QueueTrigger(Trigger):
    """A class representing queue trigger objects in the ORM."""
    __tablename__ = "Queue_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    queue_name: Mapped[str] = mapped_column(String(200))
    min_batch_size: Mapped[int]

    __mapper_args__ = {"polymorphic_identity": TriggerType.QUEUE}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = "N/A"
        return row_dict



"""This module contains a type decorator class for use in ORM models."""

import json

from sqlalchemy import Dialect, types


# pylint: disable=too-many-ancestors, abstract-method
class StringList(types.TypeDecorator):
    """A type decorator used when defining sqlalchemy columns.
    This type converts a list of strings to a json string before sending to the database.
    """
    impl = types.String
    cache_ok = True

    # pylint: disable=unused-argument
    def process_bind_param(self, value: list[str] | None, dialect: Dialect) -> str | None:
        """Convert to json string before writing to the database."""
        if value is not None:
            return json.dumps(value)

        return None

    # pylint: disable=unused-argument
    def process_result_value(self, value: str | None, dialect: Dialect) -> list[str] | None:
        """Convert from json string when retrieving from the database."""
        if value is not None:
            return json.loads(value)

        return None



"""This module handles cryptographic tasks in OpenOrchestrator."""

from cryptography.fernet import Fernet
from cryptography.exceptions import InvalidSignature


_encryption_key: str | None = None


def generate_key() -> bytes:
    """Generates a new valid AES crypto key.

    Returns:
        bytes: A valid AES crypto key.
    """
    return Fernet.generate_key()


def set_key(key: str | None) -> None:
    """Set the crypto key for the module.
    The key will be used in all subsequent calls to this module.
    """
    global _encryption_key  # pylint: disable=global-statement
    _encryption_key = key


def get_key() -> str | None:
    """Get the encryption key last set using
    crypto_util.set_key.

    Returns:
        str: The encryption key, if any.
    """
    return _encryption_key


def encrypt_string(data: str) -> str:
    """Encrypt a string using AES with the crypto key set using
    crypto_util.set_key.

    Args:
        data: The string to encrypt.

    Returns:
        str: The encrypted string.

    Raises:
        RuntimeError: If the encryption key has not been set.
    """
    if not _encryption_key:
        raise RuntimeError("Can't encrypt without an encryption key.")

    byte_data = data.encode()
    byte_data = Fernet(_encryption_key).encrypt(byte_data)
    return byte_data.decode()


def decrypt_string(data: str) -> str:
    """Decrypt a string using AES with the crypto key set using
    crypto_util.set_key.

    Args:
        data: The string to decrypt.

    Raises:
        ValueError: If the crypto key doesn't match the key that was used when encrypting.

    Returns:
        str: The decrypted string.

    Raises:
        RuntimeError: If the encryption key has not been set.
    """
    if not _encryption_key:
        raise RuntimeError("Can't decrypt without an encryption key.")

    try:
        byte_data = data.encode()
        byte_data = Fernet(_encryption_key).decrypt(byte_data)
    except InvalidSignature as exc:
        raise ValueError("Couldn't verify signature. The decryption key is not the same as the encryption key.") from exc

    return byte_data.decode()


def validate_key(key: str) -> bool:
    """Validate if a encryption key is a valid AES encryption key.

    Args:
        key: The key to validate.

    Returns:
        bool: True if the key is valid.
    """
    try:
        Fernet(key)
    except ValueError:
        return False

    return True



"""This module is the entry point for the Orchestrator app. It contains a single class
that when created starts the application."""

import socket

from nicegui import ui, app

from OpenOrchestrator.orchestrator.tabs.trigger_tab import TriggerTab
from OpenOrchestrator.orchestrator.tabs.settings_tab import SettingsTab
from OpenOrchestrator.orchestrator.tabs.logging_tab import LoggingTab
from OpenOrchestrator.orchestrator.tabs.jobs_tab import JobsTab
from OpenOrchestrator.orchestrator.tabs.constants_tab import ConstantTab
from OpenOrchestrator.orchestrator.tabs.queue_tab import QueueTab
from OpenOrchestrator.orchestrator.tabs.schedulers_tab import SchedulerTab


# pylint: disable=too-many-instance-attributes
class Application():
    """The main application of Orchestrator.
    It contains a header and the four tabs of the application.
    """
    def __init__(self, port: int | None = None, show: bool = True) -> None:
        with ui.header():
            with ui.tabs() as self.tabs:
                ui.tab('Triggers').props("auto-id=trigger_tab")
                ui.tab('Queues').props("auto-id=queues_tab")
                ui.tab('Schedulers').props("auto-id=schedulers_tab")
                ui.tab('Jobs').props("auto-id=jobs_tab")
                ui.tab('Logs').props("auto-id=logs_tab")
                ui.tab('Constants').props("auto-id=constants_tab")
                ui.tab('Settings').props("auto-id=settings_tab")

            ui.space()
            ui.button(icon="contrast", on_click=ui.dark_mode().toggle)
            ui.button(icon='refresh', on_click=self.update_tab).props("auto-id=refresh_button")

        with ui.tab_panels(self.tabs, value='Settings', on_change=self.update_tab).classes('w-full') as self.tab_panels:
            self.t_tab = TriggerTab('Triggers')
            self.j_tab = JobsTab("Jobs", self.on_job_click)
            self.l_tab = LoggingTab("Logs")
            self.q_tab = QueueTab("Queues")
            self.c_tab = ConstantTab("Constants")
            self.s_tab = SchedulerTab("Schedulers")
            SettingsTab('Settings')

        self._define_on_close()

        app.on_startup(self.update_loop)
        app.on_exception(lambda exc: ui.notify(exc, type='negative'))
        ui.run(title="Orchestrator", favicon='🤖', native=False, port=port or get_free_port(), reload=False, show=show)

    def update_tab(self):
        """Update the date in the currently selected tab."""
        match self.tab_panels.value:
            case 'Triggers':
                self.t_tab.update()
            case 'Queues':
                self.q_tab.update()
            case 'Jobs':
                self.j_tab.update()
            case 'Logs':
                self.l_tab.update()
            case 'Schedulers':
                self.s_tab.update()
            case 'Constants':
                self.c_tab.update()

    def update_loop(self):
        """Update the selected tab on a timer."""
        ui.timer(10, self.update_tab)

    def _define_on_close(self) -> None:
        """Tell the browser to ask for confirmation before leaving the page."""
        ui.add_body_html('''
            <script>
                window.addEventListener("beforeunload", (event) => event.preventDefault());
            </script>
            ''')

    def on_job_click(self, job_id: str):
        """Used by JobTab to update and switch to logging tab.

        Args:
            job_id: Job selected when switching to Logging tab.
        """
        self.l_tab.current_job_id = job_id
        self.l_tab.update()
        self.tabs.set_value('Logs')


def get_free_port():
    """Get a free port by creating a new socket and bind it
    on port 0 allowing the os to select the port.
    https://docs.python.org/3/library/socket.html#socket.create_connection

    Returns:
        A port number that should be free to use.
    """
    with socket.socket() as sock:
        sock.bind(("", 0))
        port = sock.getsockname()[1]

    return port


if __name__ in {'__main__', '__mp_main__'}:
    Application()



"""This module is responsible for the layout and functionality of the Schedulers tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.orchestrator import test_helper

COLUMNS = [
    {'name': "machine_name", 'label': "Machine Name", 'field': "Machine Name", 'align': 'left', 'sortable': True},
    {'name': "last_connection", 'label': "Last Connection", 'field': "Last Connection", 'align': 'left', 'sortable': True},
    {'name': "latest_trigger", 'label': "Latest Trigger", 'field': "Latest Trigger", 'align': 'left', 'sortable': True},
    {'name': "latest_trigger_time", 'label': "Latest Trigger Time", 'field': "Latest Trigger Time", 'align': 'left', 'sortable': True},
]


class SchedulerTab():
    """A class for the scheduler tab."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            self.schedulers_table = ui.table(title="Schedulers", columns=COLUMNS, rows=[], row_key='Machine Name', pagination=50).classes("w-full")
            self.add_column_colors()
        test_helper.set_automation_ids(self, "schedulers_tab")

    def update(self):
        """Updates the tables on the tab."""
        schedulers = db_util.get_schedulers()
        self.schedulers_table.rows = [s.to_row_dict() for s in schedulers]
        self.schedulers_table.update()

    def add_column_colors(self):
        """Add red coloring to the scheduler if more than a minute has passed since last ping."""
        self.schedulers_table.add_slot(
            "body-cell-last_connection",
            '''
            <q-td key="last_connection" :props="props">
                <q-badge v-if="(new Date() - new Date(+props.value.substr(6,4), +props.value.substr(3,2)-1, +props.value.substr(0,2), +props.value.substr(11,2), +props.value.substr(14,2))) > 60 * 1000" color='red'>
                    {{props.value}}
                </q-badge>
                <p v-else>
                    {{props.value}}
                </p>
            </q-td>
            '''
        )



"""A module for performing common tasks regarding datetimes."""


from datetime import datetime


def format_datetime(datetime_: datetime | None, default: str = 'N/A') -> str:
    """Format a datetime to a string.

    Args:
        datetime_: The datetime to format.
        default: A default string to return if the datetime is None. Defaults to 'N/A'.

    Returns:
        A datetime string in the format %d-%m-%Y %H:%M:%S.
    """
    if not datetime_:
        return default

    return datetime_.strftime("%d-%m-%Y %H:%M:%S")



"""This module contains functions to generate generic popups for various use cases."""

from nicegui import ui


async def question_popup(question: str, option1: str, option2: str, color1: str = 'primary', color2: str = 'primary') -> bool:
    """Shows a popup with a question and two buttons with the given options.
    Example:
        result = await question_popup("Do you like candy", "YES!", "Not really")

    Args:
        question: The question to display.
        option1: The text on button 1.
        option2: The text on button 2.
        color1: The color of button 1.
        color2: The color of button 2.

    Returns:
        bool: True if button 1 is clicked, or False if button 2 is clicked.
    """
    with ui.dialog(value=True).props('persistent') as dialog, ui.card():
        ui.label(question).classes("text-lg")
        with ui.row():
            ui.button(option1, on_click=lambda e: dialog.submit(True), color=color1).props("auto-id=popup_option1_button")
            ui.button(option2, on_click=lambda e: dialog.submit(False), color=color2).props("auto-id=popup_option2_button")

        return await dialog


async def info_popup(text: str) -> None:
    """Show a generic popup with the given text.

    Args:
        text: The text to display in the popup.
    """
    with ui.dialog(value=True).props('persistent') as dialog, ui.card():
        ui.label(text).classes("text-lg")
        ui.button("Close", on_click=lambda e: dialog.submit(True)).props("auto-id=popup_button")

        await dialog



"""This module is the entry point for the Scheduler app. It contains a single class
that when created starts the application."""

import tkinter
from tkinter import ttk, messagebox

from OpenOrchestrator.scheduler import settings_tab, run_tab


class Application(tkinter.Tk):
    """The main application object of the Scheduler app.
    Extends the tkinter.Tk object.
    """
    def __init__(self):
        # Disable pylint duplicate code error since it
        # mostly reacts to the layout code being similar.
        # pylint: disable=R0801
        self.running_jobs = []
        self.running = False

        super().__init__()
        self.title("OpenOrchestrator - Scheduler")
        self.geometry("850x600")
        style = ttk.Style(self)
        style.theme_use('vista')

        notebook = ttk.Notebook(self)
        notebook.pack(expand=True, fill='both')

        run_tab_ = run_tab.RunTab(notebook, self)
        self.settings_tab_ = settings_tab.SettingsTab(notebook)

        notebook.add(run_tab_, text='Run')
        notebook.add(self.settings_tab_, text="Settings")

        notebook.select(1)

        self.protocol('WM_DELETE_WINDOW', self.on_close)

        self.mainloop()

    def on_close(self):
        """Checks whether any jobs are still running and prompts the user before closing."""
        if (len(self.running_jobs) == 0
                or messagebox.askyesno('Warning', 'Some processes are still running. Closing the scheduler while processes are running will gum up the trigger tables. Are you sure you want to close?')):
            self.destroy()


if __name__ == '__main__':
    Application()



"""Contains the Base class for all ORM classes"""

from sqlalchemy.orm import DeclarativeBase


# pylint: disable=too-few-public-methods
class Base(DeclarativeBase):
    """SqlAlchemy base class for all ORM classes in this project."""



"""Database revision '526b6edac328': Add schedulers table"""

from alembic import op
import sqlalchemy as sa


# pylint: disable=invalid-name
# revision identifiers, used by Alembic.
revision: str = '526b6edac328'
down_revision = 'b67b7649b282'
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade the database."""
    op.create_table(
        'Schedulers',
        sa.Column('machine_name', sa.String(length=100), nullable=False, primary_key=True),
        sa.Column('last_update', sa.DateTime(), nullable=False),
        sa.Column('latest_trigger', sa.String(length=100), nullable=True),
        sa.Column('latest_trigger_time', sa.DateTime(), nullable=True),
    )



"""This module is responsible for the layout and functionality of the Logging tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput
from OpenOrchestrator.orchestrator import test_helper


COLUMNS = [
    {'name': "Log Time", 'label': "Log Time", 'field': "Log Time", 'align': 'left', 'sortable': True},
    {'name': "Level", 'label': "Level", 'field': "Level", 'align': 'left'},
    {'name': "Short Job ID", 'label': "Job ID", 'field': "Short Job ID", 'align': 'left', 'sortable': True},
    {'name': "Process Name", 'label': "Process Name", 'field': "Process Name", 'align': 'left', 'sortable': True},
    {'name': "Message", 'label': "Message", 'field': "Message", 'align': 'left', ':format': 'value => value.length < 100 ? value : value.substring(0, 100)+"..."'},
    {'name': "Full Job ID", 'label': "Full Job ID", 'field': "Full Job ID", 'headerClasses': 'hidden', 'classes': 'hidden'},
    {'name': "ID", 'label': "ID", 'field': "ID", 'headerClasses': 'hidden', 'classes': 'hidden'}
]


# pylint: disable-next=too-few-public-methods, too-many-instance-attributes
class LoggingTab():
    """The 'Logs' tab object."""
    def __init__(self, tab_name: str) -> None:
        self.current_job_id: str | None = None
        self.order_by = "Log Time"
        self.order_descending = True
        self.page = 1
        self.rows_per_page = 25
        self.log_count = 0

        with ui.tab_panel(tab_name):
            with ui.row().classes("w-full justify-between"):
                with ui.row():
                    self.from_input = DatetimeInput("From Date", on_change=self.update, allow_empty=True)
                    self.to_input = DatetimeInput("To Date", on_change=self.update, allow_empty=True)
                    self.level_input = ui.select(["All", "Trace", "Info", "Error"], value="All", label="Level", on_change=self.update).classes("w-48")
                    self.process_input = ui.select(["All"], label="Process Name", value="All", on_change=self.update).classes("w-48")
                    with ui.column().classes("items-end") as self.job_filter_container:
                        self.job_filter_label = ui.label("")
                        self.all_jobs_button = ui.button("Show all jobs", on_click=self._show_all_jobs)

                self.logs_table = ui.table(title="Logs", columns=COLUMNS, rows=[], row_key='ID',
                                           pagination={'rowsPerPage': self.rows_per_page,
                                                       'rowsNumber': self.log_count})
                self.logs_table.classes("w-full sticky-header h-[calc(100vh-200px)] overflow-auto")
                self.logs_table.props(":rows-per-page-options='[10, 25, 50, 100, 1000]' rows-per-page-label='Logs per page:'")
                self.logs_table.on("rowClick", self._row_click)
                self.logs_table.on('request', self._on_table_request)

            test_helper.set_automation_ids(self, "logs_tab")

    def update(self):
        """Update the logs table and Process input list"""
        self._update_table()
        self._update_process_input()
        self._update_job_filter_display()

    def _update_job_filter_display(self):
        """Update visibility and text of job filter."""
        if self.current_job_id:
            self.job_filter_label.set_text(f"Filtering by Job: {self.current_job_id}...")
            self.job_filter_container.set_visibility(True)
        else:
            self.job_filter_container.set_visibility(False)

    def _update_table(self):
        """Update the table with logs from the database applying the filters."""
        from_date = self.from_input.get_datetime()
        to_date = self.to_input.get_datetime()
        level = LogLevel(self.level_input.value) if self.level_input.value != "All" else None
        process_name = self.process_input.value if self.process_input.value != 'All' else None

        offset = (self.page - 1) * self.rows_per_page
        order_by = str(self.order_by).lower().replace(" ", "_")

        logs, count = db_util.get_logs(offset, limit=self.rows_per_page, from_date=from_date, to_date=to_date, log_level=level, process_name=process_name, job_id=self.current_job_id, order_by=order_by, order_desc=self.order_descending, include_count=True)
        self._update_pagination(count)
        self.logs_table.update_rows([log.to_row_dict() for log in logs])

    def _on_table_request(self, e):
        """Called when updating table pagination and sorting, to handle these manually and allow for server side pagination.

        Args:
            e: The event triggering the request.
        """
        pagination = e.args['pagination']
        self.page = pagination.get('page')
        self.rows_per_page = pagination.get('rowsPerPage')
        self.order_by = pagination.get('sortBy')
        self.order_descending = pagination.get('descending', False)
        self._update_table()

    def _update_pagination(self, log_count):
        """Update pagination element.

        Args:
            log_count: The element count of the current filtered table.
        """
        self.log_count = log_count
        self.logs_table.pagination = {"rowsNumber": self.log_count,
                                      "page": self.page,
                                      "rowsPerPage": self.rows_per_page,
                                      "sortBy": self.order_by,
                                      "descending": self.order_descending}
        self.logs_table.update()

    def _update_process_input(self):
        """Update the process input with names from the database."""
        process_names = list(db_util.get_unique_log_process_names())
        process_names.insert(0, "All")
        self.process_input.options = process_names
        self.process_input.update()

    def set_job_filter(self, job_id: str | None):
        """Set filter to specific job."""
        self.current_job_id = job_id
        self.update()

    def _show_all_jobs(self):
        self.set_job_filter(None)

    def _row_click(self, event):
        """Display a dialog with info on the clicked log."""
        row = event.args[1]
        with ui.dialog(value=True) as dialog, ui.card():
            ui.label("Log ID:").classes("font-bold")
            ui.label(row['ID'])
            ui.label("Log Time:").classes("font-bold")
            ui.label(row['Log Time'])
            ui.label("Process Name:").classes("font-bold")
            ui.label(row['Process Name'])
            ui.label("Log Level:").classes("font-bold")
            ui.label(row['Level'])
            ui.label("Message:").classes("font-bold")
            ui.html(f"<pre>{row['Message']}</pre>")
            ui.label("Job ID:").classes("font-bold")
            ui.label(row['Full Job ID'])
            with ui.row():
                ui.button("Show job logs", on_click=lambda: [self.set_job_filter(row['Full Job ID']), dialog.close()]).set_enabled(row['Full Job ID'] is not None)
                ui.button("Close", on_click=dialog.close)



"""This module tests the OpenOrchestrator.scheduler.runner module."""

import unittest
from unittest.mock import patch, MagicMock
from datetime import datetime
import subprocess

from OpenOrchestrator.scheduler import runner
from OpenOrchestrator.database import db_util
from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database.triggers import TriggerStatus, SingleTrigger
from OpenOrchestrator.database.jobs import JobStatus
from OpenOrchestrator.database.logs import LogLevel


TEST_MODULE = "OpenOrchestrator.scheduler.runner"


class TestSchedulerRunner(unittest.TestCase):
    """Test the runner functionality of the scheduler."""
    def setUp(self) -> None:
        db_test_util.establish_clean_database()

    @patch(f"{TEST_MODULE}.shutil.which", return_value="git")
    @patch(f"{TEST_MODULE}.os.makedirs")
    @patch(f"{TEST_MODULE}.subprocess.run")
    @patch(f"{TEST_MODULE}.get_repo_folder_path", return_value="repo_folder")
    def test_clone_git_repo(self, mock_get_repo_folder_path: MagicMock, mock_run: MagicMock, mock_makedirs: MagicMock, mock_which: MagicMock):
        """Test the clone_git_repo function of the runner module.

        Args:
            mock_get_repo_folder_path: A MagicMock of the runner.get_repo_folder_path.
            mock_run: A MagicMock of the subprocess.run function.
            mock_makedirs: A MagicMock of the os.makedirs function.
            mock_which: A MagicMock of the shutil.which function.
        """
        repo_url = "https://mock_repo.com/git"

        # Run without branch
        repo_path = runner.clone_git_repo(repo_url, "")

        self.assertRegex(repo_path, r"repo_folder\\[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}")
        mock_get_repo_folder_path.assert_called_once()
        mock_makedirs.assert_called_once_with(repo_path)
        mock_which.assert_called_once_with("git")
        mock_run.assert_called_once_with(["git", "clone", repo_url, repo_path], check=True)

        # Run with branch
        mock_run.reset_mock()
        repo_path = runner.clone_git_repo(repo_url, "branch1")
        mock_run.assert_called_once_with(["git", "clone", "-b", "branch1", repo_url, repo_path], check=True)

    @patch(f"{TEST_MODULE}.util.get_scheduler_name", return_value="Machine Name")
    @patch(f"{TEST_MODULE}.subprocess.Popen")
    @patch(f"{TEST_MODULE}.os.path.isfile", return_value=True)
    @patch(f"{TEST_MODULE}.find_main_file", return_value="main.py")
    @patch(f"{TEST_MODULE}.clone_git_repo", return_value="folder_path")
    def test_run_trigger(self, mock_clone_git_repo: MagicMock, mock_find_main_file: MagicMock, mock_isfile: MagicMock,
                         mock_popen: MagicMock, mock_get_scheduler_name: MagicMock):
        """Test the run_trigger function in the runner module.

        Args:
            mock_clone_git_repo: A MagicMock of the runner.clone_git_repo function.
            mock_find_main_file: A MagicMock of the runner.find_main_file function.
            mock_isfile: A MagicMock of the os.path.isfile function.
            mock_Popen: A MagicMock of the subprocess.Popen class.
            mock_get_scheduler_name: A MagicMock of the util.get_scheduler_name function.
        """
        trigger_id = db_util.create_single_trigger(
            trigger_name="Trigger Name",
            process_name="Process Name",
            next_run=datetime.now(),
            process_path="https://mock_repo.com/git",
            process_args="Arguments",
            is_git_repo=True,
            is_blocking=False,
            priority=0,
            scheduler_whitelist=[],
            git_branch="Branch"
        )

        trigger = db_util.get_trigger(trigger_id)

        # No jobs before
        jobs_before = db_util.get_jobs()
        self.assertEqual(len(jobs_before), 0)

        scheduler_job = runner.run_trigger(trigger)

        # Check the scheduler_job object
        self.assertEqual(scheduler_job.trigger, trigger)
        self.assertEqual(scheduler_job.process_folder, "folder_path")

        # Job should be created
        jobs_after = db_util.get_jobs()
        self.assertEqual(len(jobs_after), 1)

        job = jobs_after[0]
        self.assertEqual(job.process_name, "Process Name")
        self.assertEqual(job.scheduler_name, "Machine Name")
        self.assertEqual(job.status, JobStatus.RUNNING)
        self.assertIsNotNone(job.start_time)
        self.assertIsNone(job.end_time)

        # Verify job is in SchedulerJob
        self.assertEqual(scheduler_job.job.id, job.id)

        # Check that the mock functions were called correctly
        mock_clone_git_repo.assert_called_once_with(trigger.process_path, trigger.git_branch)
        mock_find_main_file.assert_called_once_with("folder_path")
        mock_isfile.assert_called_once_with("main.py")
        mock_popen.assert_called_once_with(['python', "main.py", trigger.process_name, db_util.get_conn_string(), crypto_util.get_key(), trigger.process_args, str(trigger.id), str(scheduler_job.job.id)],
                                           stderr=subprocess.PIPE, text=True)
        mock_get_scheduler_name.assert_called_once()

        # Check that trigger status was set
        trigger = db_util.get_trigger(trigger_id)
        self.assertEqual(trigger.process_status, TriggerStatus.RUNNING)

        # Check that a ping was sent to the schedulers list
        schedulers = db_util.get_schedulers()
        self.assertEqual(schedulers[0].machine_name, "Machine Name")
        self.assertEqual(schedulers[0].latest_trigger, "Trigger Name")

    def test_end_job_updates_status(self):
        """Test that end_job sets job status to DONE and clears process folder."""
        # Create a job manually
        job_obj = db_util.start_job("Process", "Scheduler")

        # Create mock SchedulerJob
        mock_process = MagicMock()
        trigger_id = db_util.create_single_trigger(
            trigger_name="Trigger Name",
            process_name="Process Name",
            next_run=datetime.now(),
            process_path="https://mock_repo.com/git",
            process_args="Arguments",
            is_git_repo=True,
            is_blocking=False,
            priority=0,
            scheduler_whitelist=[],
            git_branch="Branch"
        )

        trigger = db_util.get_trigger(trigger_id)

        scheduler_job = runner.SchedulerJob(
            process=mock_process,
            trigger=trigger,
            job=job_obj,
            process_folder="test_folder"
        )

        # End the job
        with patch(f"{TEST_MODULE}.clear_folder") as mock_clear:
            runner.end_job(scheduler_job)

        # Verify job status updated
        updated_job = db_util.get_job(job_obj.id)
        self.assertEqual(updated_job.status, JobStatus.DONE)
        self.assertIsNotNone(updated_job.end_time)

        # Verify folder cleared
        mock_clear.assert_called_once_with("test_folder")

    def test_fail_job_updates_status(self):
        """Test that fail_job sets job status to FAILED and creates error log."""
        job_obj = db_util.start_job("Process", "Scheduler")

        # Create mock SchedulerJob
        mock_process = MagicMock()
        mock_process.communicate.return_value = ("", "Error message from stderr")

        trigger = MagicMock(spec=SingleTrigger)
        trigger_id = db_util.create_single_trigger(
            trigger_name="Trigger Name",
            process_name="Process Name",
            next_run=datetime.now(),
            process_path="https://mock_repo.com/git",
            process_args="Arguments",
            is_git_repo=True,
            is_blocking=False,
            priority=0,
            scheduler_whitelist=[],
            git_branch="Branch"
        )

        trigger = db_util.get_trigger(trigger_id)

        scheduler_job = runner.SchedulerJob(
            process=mock_process,
            trigger=trigger,
            job=job_obj,
            process_folder=None
        )

        # Fail the job
        runner.fail_job(scheduler_job)

        # Verify job status
        updated_job = db_util.get_job(job_obj.id)
        self.assertEqual(updated_job.status, JobStatus.FAILED)
        self.assertIsNotNone(updated_job.end_time)

        # Verify error log created
        logs = db_util.get_logs(offset=0, limit=1)
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0].log_level, LogLevel.ERROR)
        self.assertIn("Error message from stderr", logs[0].log_message)

    def test_kill_job_updates_status(self):
        """Test that kill_job sets job status to KILLED and kills process."""
        job_obj = db_util.start_job("Process", "Scheduler")

        # Create mock SchedulerJob
        mock_process = MagicMock()
        trigger_id = db_util.create_single_trigger(
            trigger_name="Trigger Name",
            process_name="Process Name",
            next_run=datetime.now(),
            process_path="https://mock_repo.com/git",
            process_args="Arguments",
            is_git_repo=True,
            is_blocking=False,
            priority=0,
            scheduler_whitelist=[],
            git_branch="Branch"
        )

        trigger = db_util.get_trigger(trigger_id)

        scheduler_job = runner.SchedulerJob(
            process=mock_process,
            trigger=trigger,
            job=job_obj,
            process_folder=None
        )

        # Kill the job
        runner.kill_job(scheduler_job)

        # Verify process was killed
        mock_process.kill.assert_called_once()

        # Verify job status
        updated_job = db_util.get_job(job_obj.id)
        self.assertEqual(updated_job.status, JobStatus.KILLED)
        self.assertIsNotNone(updated_job.end_time)


if __name__ == '__main__':
    unittest.main()



"""This module contains database utility functions used in testing."""

from datetime import datetime, timedelta
import os

from sqlalchemy import text
from OpenOrchestrator.database import db_util, base
from OpenOrchestrator.common import crypto_util


def establish_clean_database():
    """Connect to the database, drop all tables and recreate them."""
    db_util.connect(os.environ["CONN_STRING"])
    crypto_util.set_key(crypto_util.generate_key().decode())

    drop_all_tables()
    base.Base.metadata.create_all(db_util._connection_engine)  # pylint: disable=protected-access


def drop_all_tables():
    """Drop all ORM tables from the database."""
    engine = db_util._connection_engine  # pylint: disable=protected-access
    if not engine:
        raise RuntimeError("Not connected to a database.")

    base.Base.metadata.drop_all(engine)

    # alembic_version is not part of the ORM, drop it seperately.
    with engine.connect() as conn:
        conn.execute(text("DROP TABLE IF EXISTS alembic_version"))
        conn.commit()


def reset_triggers():
    """Delete all triggers in the database and create a new of each."""
    for t in db_util.get_all_triggers():
        db_util.delete_trigger(t.id)

    if len(db_util.get_all_triggers()) != 0:
        raise RuntimeError("Not all triggers were deleted.")

    next_run = datetime.now() - timedelta(seconds=2)
    db_util.create_single_trigger("Single", "Process1", next_run, "Path", "Args", False, False, 0)
    db_util.create_scheduled_trigger("Scheduled", "Process1", "0 0 * * *", next_run, "Path", "Args", False, False, 0)
    db_util.create_queue_trigger("Queue", "Process1", "Trigger Queue", "Path", "Args", False, False, 2, 0)



"""This module is responsible for the layout and functionality of the 'New constant' popup."""

from __future__ import annotations
from typing import TYPE_CHECKING

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.constants import Constant
from OpenOrchestrator.orchestrator.popups.generic_popups import question_popup
from OpenOrchestrator.orchestrator import test_helper

if TYPE_CHECKING:
    from OpenOrchestrator.orchestrator.tabs.constants_tab import ConstantTab


# pylint: disable-next=too-few-public-methods, too-many-instance-attributes
class ConstantPopup():
    """A popup for creating/updating queue triggers."""
    def __init__(self, constant_tab: ConstantTab, constant: Constant | None = None):
        """Create a new popup.
        If a constant is given it will be updated instead of creating a new constant.

        Args:
            constant: The constant to update if any.
        """
        self.constant_tab = constant_tab
        self.constant = constant
        title = 'Update Constant' if constant else 'New Constant'
        button_text = "Update" if constant else "Create"

        with ui.dialog(value=True).props('persistent') as self.dialog, ui.card().classes('w-full'):
            ui.label(title).classes("text-xl")
            self.name_input = ui.input("Constant Name").classes("w-full")
            self.value_input = ui.input("Constant Value").classes("w-full")

            with ui.row():
                self.save_button = ui.button(button_text, on_click=self._create_constant)
                self.cancel_button = ui.button("Cancel", on_click=self.dialog.close)

                if constant:
                    self.delete_button = ui.button("Delete", color='negative', on_click=self._delete_constant)

        self._define_validation()
        self._pre_populate()
        test_helper.set_automation_ids(self, "constant_popup")

    def _define_validation(self):
        """Define validation rules for input elements."""
        self.name_input._validation = {"Please enter a name": bool}  # pylint: disable=protected-access
        self.value_input._validation = {"Please enter a value": bool}  # pylint: disable=protected-access

    def _pre_populate(self):
        """Pre populate the inputs with an existing constant."""
        if self.constant:
            self.name_input.value = self.constant.name
            self.name_input.disable()
            self.value_input.value = self.constant.value

    def _create_constant(self):
        """Creates a new constant in the database using the data from the
        UI.
        """
        self.name_input.validate()
        self.value_input.validate()

        if self.name_input.error or self.value_input.error:
            return

        name = self.name_input.value
        value = self.value_input.value

        if self.constant:
            db_util.update_constant(name, value)
        else:
            # Check if constant already exists
            try:
                db_util.get_constant(name)
                exists = True
            except ValueError:
                exists = False

            if exists:
                ui.notify("A constant with that name already exists.", type='negative')
                return

            db_util.create_constant(name, value)

        self.dialog.close()
        self.constant_tab.update()

    async def _delete_constant(self):
        if not self.constant:
            return
        if await question_popup(f"Delete constant '{self.constant.name}?", "Delete", "Cancel", color1='negative'):
            db_util.delete_constant(self.constant.name)
            self.dialog.close()
            self.constant_tab.update()



"""This module is responsible for the layout and functionality of the Schedulers tab
in Orchestrator."""
from typing import Callable

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.orchestrator import test_helper

COLUMNS = [
    {'name': "job_id", 'label': "ID", 'field': "ID", 'headerClasses': 'hidden', 'classes': 'hidden'},
    {'name': "process_name", 'label': "Process Name", 'field': "Process Name", 'align': 'left', 'sortable': True},
    {'name': "scheduler_name", 'label': "Scheduler", 'field': "Scheduler", 'align': 'left', 'sortable': True},
    {'name': "start_time", 'label': "Start Time", 'field': "Start Time", 'align': 'left', 'sortable': True},
    {'name': "end_time", 'label': "End Time", 'field': "End Time", 'align': 'left', 'sortable': True},
    {'name': "status", 'label': "Status", 'field': "Status", 'align': 'left', 'sortable': True}
]


# pylint: disable-next=too-few-public-methods
class JobsTab:
    """A class for the jobs tab."""
    def __init__(self, tab_name: str, on_job_click: Callable[[str], None]) -> None:
        with ui.tab_panel(tab_name):
            self.jobs_table = ui.table(title="Jobs", columns=COLUMNS, rows=[], row_key='job_id', pagination=50).classes("w-full")
            self.jobs_table.on("rowClick", self._row_click)
            self._add_column_colors()
        test_helper.set_automation_ids(self, "jobs_tab")
        self.on_job_click = on_job_click

    def update(self):
        """Updates the tables on the tab."""
        jobs = db_util.get_jobs()
        self.jobs_table.rows = [s.to_row_dict() for s in jobs]

    def _row_click(self, event):
        row = event.args[1]
        job_id = row["ID"]
        self.on_job_click(job_id)

    def _add_column_colors(self):
        """Add custom coloring to the jobs table."""
        # Add coloring to the status column
        color_dict = "{Running: 'blue', Done: 'green', Failed: 'red', Killed: 'grey-9'}"

        self.jobs_table.add_slot(
            "body-cell-status",
            f'''
            <q-td key="status" :props="props">
                <q-badge v-if="{color_dict}[props.value]" :color="{color_dict}[props.value]">
                    {{{{props.value}}}}
                </q-badge>
                <p v-else>
                    {{{{props.value}}}}
                </p>
            </q-td>
            '''
        )



"""Tests relating to the logs tab in Orchestrator."""

import unittest
from datetime import datetime, timedelta
import time

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.tests.ui_tests import ui_util


class TestLogsTab(unittest.TestCase):
    """Test functionality of the logs tab ui."""
    def setUp(self) -> None:
        self.browser = ui_util.open_orchestrator()
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab]").click()
        ui_util.refresh_ui(self.browser)

    def tearDown(self) -> None:
        self.browser.quit()

    @ui_util.screenshot_on_error
    def test_logs_table(self):
        """Test that logs are shown correctly in the logs table."""
        # Create some logs
        self._create_logs()

        ui_util.refresh_ui(self.browser)

        # Check result
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertRegex(table_data[0][0], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(table_data[0][1], LogLevel.ERROR.value)
        self.assertEqual(table_data[0][3], "Test Error")
        self.assertEqual(table_data[0][4], "Error Message")

        self.assertRegex(table_data[1][0], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(table_data[1][1], LogLevel.INFO.value)
        self.assertEqual(table_data[1][3], "Test Info")
        self.assertEqual(table_data[1][4], "Info Message")

        self.assertRegex(table_data[2][0], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")
        self.assertEqual(table_data[2][1], LogLevel.TRACE.value)
        self.assertEqual(table_data[2][3], "Test Trace")
        self.assertEqual(table_data[2][4], "Trace Message")

    @ui_util.screenshot_on_error
    def test_date_filter(self):
        """Test filtering on date."""
        yesterday = datetime.today() - timedelta(days=1)
        tomorrow = datetime.today() + timedelta(days=1)

        self._create_logs()
        ui_util.refresh_ui(self.browser)

        # No filter
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

        # From yesterday
        self._set_date_filter(from_date=yesterday, to_date=None)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

        # From tomorrow
        self._set_date_filter(from_date=tomorrow, to_date=None)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 0)

        # To yesterday
        self._set_date_filter(from_date=None, to_date=yesterday)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 0)

        # To tomorrow
        self._set_date_filter(from_date=None, to_date=tomorrow)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

        # From yesterday To tomorrow
        self._set_date_filter(from_date=yesterday, to_date=tomorrow)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

        # Clear filter
        self._set_date_filter(None, None)

    @ui_util.screenshot_on_error
    def test_process_filter(self):
        """Test filtering on process name."""
        self._create_logs()

        self._set_process_filter(2)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Error")

        self._set_process_filter(3)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Info")

        self._set_process_filter(4)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Trace")

        self._set_process_filter(1)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

    @ui_util.screenshot_on_error
    def test_level_filter(self):
        """Test filtering on log level."""
        self._create_logs()

        self._set_level_filter(2)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Trace")

        self._set_level_filter(3)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Info")

        self._set_level_filter(4)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(table_data[0][3], "Test Error")

        self._set_level_filter(1)
        table_data = ui_util.get_table_data(self.browser, "logs_tab_logs_table")
        self.assertEqual(len(table_data), 3)

    def _set_date_filter(self, from_date: datetime | None, to_date: datetime | None):
        # Clear filters
        from_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab_from_input]")
        from_input.send_keys(Keys.CONTROL, "a", Keys.DELETE)
        to_input = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab_to_input]")
        to_input.send_keys(Keys.CONTROL, "a", Keys.DELETE)

        if from_date:
            from_input.send_keys(from_date.strftime("%d-%m-%Y %H:%M:%S"))

        if to_date:
            to_input.send_keys(to_date.strftime("%d-%m-%Y %H:%M:%S"))

    def _set_process_filter(self, index: int):
        """Select a process in the process filter.

        Args:
            index: The 1-based index of the process to choose.
        """
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab_process_input]").click()
        self.browser.find_element(By.XPATH, f"//div[@role='listbox']//div[contains(@class, 'q-item')][{index}]").click()
        time.sleep(0.5)

    def _set_level_filter(self, index: int):
        """Select a level in the level filter.

        Args:
            index: The 1-based index of the level to choose.
        """
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab_level_input]").click()
        self.browser.find_element(By.XPATH, f"//div[@role='listbox']//div[contains(@class, 'q-item')][{index}]").click()
        time.sleep(0.5)

    def _create_logs(self):
        """Create some logs for testing."""
        db_util.create_log("Test Trace", LogLevel.TRACE, None, "Trace Message")
        time.sleep(0.1)
        db_util.create_log("Test Info", LogLevel.INFO, None, "Info Message")
        time.sleep(0.1)
        db_util.create_log("Test Error", LogLevel.ERROR, None, "Error Message")

        ui_util.refresh_ui(self.browser)


if __name__ == '__main__':
    unittest.main()



"""This module contains a single class called OrchestratorConnection.
The easiest way to create an OrchestratorConnection object is to call the
class method create_connection_from_args."""

import sys
from datetime import datetime

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.queues import QueueElement, QueueStatus
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.database.constants import Constant, Credential
from OpenOrchestrator.database.triggers import TriggerStatus


class OrchestratorConnection:
    """An OrchestratorConnection is used to easier communicate with
    OpenOrchestrator within a running process. If used in conjunction with
    OpenOrchestrator's Scheduler use orchestrator_connection.create_connection_from_args
    to instead of initializing the object manually.
    """

    def __init__(self, process_name: str, connection_string: str, crypto_key: str, process_arguments: str, trigger_id: str, job_id: str):
        """
        Args:
            process_name: A human friendly tag to identify the process.
            connection_string: An ODBC connection to the OpenOrchestrator database.
            crypto_key: Secret key for decrypting database content.
            process_arguments (optional): Arguments for the controlling how the process should run.
            trigger_id: ID of trigger used to start this process.
            job_id: ID of job created when starting this process.
        """
        self.process_name = process_name
        self.process_arguments = process_arguments
        self.trigger_id = trigger_id
        self.job_id = job_id
        crypto_util.set_key(crypto_key)
        db_util.connect(connection_string)

    def __repr__(self):
        return f"OrchestratorConnection - Process name: {self.process_name}"

    def log_trace(self, message: str) -> None:
        """Create a message in the Orchestrator log with a level of 'trace'.
        The log is automatically annotated with the current time and name of the process
        Args:
            message: Message to be logged.
        """
        db_util.create_log(self.process_name, LogLevel.TRACE, self.job_id, message)

    def log_info(self, message: str) -> None:
        """Create a message in the Orchestrator log with a level of 'info'.
        The log is automatically annotated with the current time and name of the process
        Args:
            message: Message to be logged.
        """
        db_util.create_log(self.process_name, LogLevel.INFO, self.job_id, message)

    def log_error(self, message: str) -> None:
        """Create a message in the Orchestrator log with a level of 'error'.
        The log is automatically annotated with the current time and name of the process
        Args:
            message: Message to be logged.
        """
        db_util.create_log(self.process_name, LogLevel.ERROR, self.job_id, message)

    def get_constant(self, constant_name: str) -> Constant:
        """Get a constant from the database.

        Args:
            name: The name of the constant.

        Returns:
            Constant: The constant with the given name.

        Raises:
            ValueError: If no constant with the given name exists.
        """
        return db_util.get_constant(constant_name)

    def get_credential(self, credential_name: str) -> Credential:
        """Get a credential from the database.
        The password of the credential is decrypted.

        Args:
            name: The name of the credential.

        Returns:
            Credential: The credential with the given name.

        Raises:
            ValueError: If no credential with the given name exists.
        """
        return db_util.get_credential(credential_name)

    def update_constant(self, constant_name: str, new_value: str) -> None:
        """Updates an existing constant with a new value.

        Args:
            name: The name of the constant to update.
            new_value: The new value of the constant.
        """
        db_util.update_constant(constant_name, new_value)

    def update_credential(self, credential_name: str, new_username: str, new_password: str) -> None:
        """Updates an existing credential with a new value.

        Args:
            name: The name of the credential to update.
            new_username: The new username of the credential.
            new_password: The new password of the credential.
        """
        db_util.update_credential(credential_name, new_username, new_password)

    def create_queue_element(self, queue_name: str, reference: str | None = None, data: str | None = None, created_by: str | None = None) -> QueueElement:
        """Adds a queue element to the given queue.

        Args:
            queue_name: The name of the queue to add the element to.
            reference (optional): The reference of the queue element.
            data (optional): The data of the queue element.
            created_by (optional): The name of the creator of the queue element.

        Returns:
            QueueElement: The created queue element.
        """
        return db_util.create_queue_element(queue_name, reference, data, created_by)

    def bulk_create_queue_elements(self, queue_name: str, references: tuple[str | None, ...], data: tuple[str | None, ...],
                                   created_by: str | None = None) -> None:
        """Insert multiple queue elements into a queue in an optimized manner.
        The lengths of both 'references' and 'data' must be equal to the number of elements to insert.

        Args:
            queue_name: The name of the queue to insert into.
            references: A tuple of reference strings for each queue element.
            data: A tuple of data strings for each queue element.
            created_by (Optional): The name of the creator of the queue elements.

        Raises:
            ValueError: If either 'references' or 'data' are empty, or if they are not equal in length.
        """
        db_util.bulk_create_queue_elements(queue_name, references, data, created_by)

    def get_next_queue_element(self, queue_name: str, reference: str | None = None,
                               set_status: bool = True) -> QueueElement | None:
        """Gets the next queue element from the given queue that has the status 'new'.

        Args:
            queue_name: The name of the queue to retrieve from.
            reference (optional): The reference to filter on. If None the filter is disabled.
            set_status (optional): If true the queue element's status is set to 'in progress'
                and the start time is noted.

        Returns:
            QueueElement | None: The next queue element in the queue if any.
        """
        return db_util.get_next_queue_element(queue_name, reference, set_status)

    def get_queue_elements(self, queue_name: str, reference: str | None = None, status: QueueStatus | None = None,
                           offset: int = 0, limit: int = 100, from_date: datetime | None = None, to_date: datetime | None = None) -> tuple[QueueElement, ...]:
        """Get multiple queue elements from a queue. The elements are ordered by created_date.

        Args:
            queue_name: The queue to get elements from.
            reference (optional): The reference to filter by. If None the filter is disabled.
            status (optional): The status to filter by if any. If None the filter is disabled.
            offset: The number of queue elements to skip.
            limit: The number of queue elements to get.

        Returns:
            tuple[QueueElement]: A tuple of queue elements.
        """
        return db_util.get_queue_elements(queue_name, reference, status, from_date, to_date, offset, limit)

    def set_queue_element_status(self, element_id: str, status: QueueStatus, message: str | None = None) -> None:
        """Set the status of a queue element.
        If the new status is 'in progress' the start date is noted.
        If the new status is 'Done', 'Failed' or 'Abandoned' the end date is noted.

        Args:
            element_id: The id of the queue element to change status on.
            status: The new status of the queue element.
            message (Optional): The message to attach to the queue element. This overrides any existing messages.
        """
        db_util.set_queue_element_status(element_id, status, message)

    def delete_queue_element(self, element_id: str) -> None:
        """Delete a queue element from the database.

        Args:
            element_id: The id of the queue element.
        """
        db_util.delete_queue_element(element_id)

    def is_trigger_pausing(self) -> bool:
        """Check if my trigger is pausing.

        Returns:
            bool: Whether or not the trigger used to start this process is pausing.
        """
        my_trigger = db_util.get_trigger(self.trigger_id)
        return my_trigger.process_status in (TriggerStatus.PAUSING, TriggerStatus.PAUSED)

    def pause_my_trigger(self) -> None:
        """Pause the trigger used to start this process."""
        db_util.set_trigger_status(self.trigger_id, TriggerStatus.PAUSING)

    @classmethod
    def create_connection_from_args(cls):
        """Create a Connection object using the arguments passed to sys.argv.
        This function is the preferred way to create a connection with OpenOrchestrator's Scheduler"""
        process_name = sys.argv[1]
        connection_string = sys.argv[2]
        crypto_key = sys.argv[3]
        process_arguments = sys.argv[4]
        trigger_id = sys.argv[5]
        job_id = sys.argv[6]
        return OrchestratorConnection(process_name, connection_string, crypto_key, process_arguments, trigger_id, job_id)



"""This module defines ORM classes for jobs"""

from datetime import datetime
import enum
import uuid

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class JobStatus(enum.Enum):
    """An enum representing the level of logs."""
    RUNNING = "Running"
    DONE = "Done"
    FAILED = "Failed"
    KILLED = "Killed"


class Job(Base):
    """A class representing job objects in the ORM."""

    __tablename__ = "Jobs"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    process_name: Mapped[str] = mapped_column(String(100))
    scheduler_name: Mapped[str] = mapped_column(String(100))
    status: Mapped[JobStatus] = mapped_column(default=JobStatus.RUNNING)
    start_time: Mapped[datetime] = mapped_column(default=datetime.now)
    end_time: Mapped[datetime] = mapped_column(default=None, nullable=True)

    def to_row_dict(self) -> dict[str, str]:
        """Convert log to a row dictionary for display in a table."""
        return {
            "ID": str(self.id),
            "Process Name": self.process_name,
            "Scheduler": self.scheduler_name,
            "Start Time":  datetime_util.format_datetime(self.start_time),
            "End Time":  datetime_util.format_datetime(self.end_time),
            "Status": self.status.value
        }



"""Tests relating to the constants tab in Orchestrator."""

import unittest
import time

from selenium.webdriver.common.by import By

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.tests.ui_tests import ui_util


class TestConstantsTab(unittest.TestCase):
    """Test functionality of the constants tab ui."""
    def setUp(self) -> None:
        self.browser = ui_util.open_orchestrator()
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constants_tab]").click()
        ui_util.refresh_ui(self.browser)

    def tearDown(self) -> None:
        self.browser.quit()

    @ui_util.screenshot_on_error
    def test_constants_table(self):
        """Test that constants are shown correctly in the constants table."""
        db_util.create_constant("Constant 1", "Value 1")
        db_util.create_constant("Constant 2", "Value 2")
        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "constants_tab_constants_table")

        self.assertEqual(table_data[0][0], "Constant 1")
        self.assertEqual(table_data[0][1], "Value 1")
        self.assertRegex(table_data[0][2], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")

        self.assertEqual(table_data[1][0], "Constant 2")
        self.assertEqual(table_data[1][1], "Value 2")
        self.assertRegex(table_data[1][2], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")

    @ui_util.screenshot_on_error
    def test_credentials_table(self):
        """Test that credentials are shown correctly in the credentials table."""
        db_util.create_credential("Credential 1", "Username 1", "abc")
        db_util.create_credential("Credential 2", "Username 2", "abcdefghijklmnopq")
        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "constants_tab_credentials_table")

        self.assertEqual(table_data[0][0], "Credential 1")
        self.assertEqual(table_data[0][1], "Username 1")
        self.assertEqual(table_data[0][2], "100 encrypted bytes. 0-15 decrypted bytes.")
        self.assertRegex(table_data[0][3], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")

        self.assertEqual(table_data[1][0], "Credential 2")
        self.assertEqual(table_data[1][1], "Username 2")
        self.assertEqual(table_data[1][2], "120 encrypted bytes. 16-31 decrypted bytes.")
        self.assertRegex(table_data[1][3], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")

    @ui_util.screenshot_on_error
    def test_create_constant(self):
        """Test creating a constant."""
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constants_tab_constant_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_name_input]").send_keys("Constant Name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_value_input]").send_keys("Constant Value")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_save_button]").click()
        time.sleep(1)

        constant = db_util.get_constant("Constant Name")
        self.assertEqual(constant.name, "Constant Name")
        self.assertEqual(constant.value, "Constant Value")

    @ui_util.screenshot_on_error
    def test_create_credential(self):
        """Test creating a credential."""
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constants_tab_credential_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_name_input]").send_keys("Credential Name")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_username_input]").send_keys("Credential Username")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_password_input]").send_keys("Credential Password")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_save_button]").click()
        time.sleep(1)

        # Make sure the encryption key matches the one in the ui
        crypto_util.set_key(ui_util.ENCRYPTION_KEY)

        credential = db_util.get_credential("Credential Name")
        self.assertEqual(credential.name, "Credential Name")
        self.assertEqual(credential.username, "Credential Username")
        self.assertEqual(credential.password, "Credential Password")

    @ui_util.screenshot_on_error
    def test_edit_constant(self):
        """Test editing an existing constant."""
        # Create a constant
        db_util.create_constant("Constant Name", "Value")
        ui_util.refresh_ui(self.browser)

        # Click constant in table
        ui_util.click_table_row(self.browser, "constants_tab_constants_table", 0)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_value_input]").send_keys(" New")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_save_button]").click()
        time.sleep(1)

        constant = db_util.get_constant("Constant Name")
        self.assertEqual(constant.value, "Value New")

    @ui_util.screenshot_on_error
    def test_edit_credential(self):
        """Test editing an existing credential."""
        # Create a credential
        db_util.create_credential("Credential Name", "Username", "Password")
        ui_util.refresh_ui(self.browser)

        # Click credential in table
        ui_util.click_table_row(self.browser, "constants_tab_credentials_table", 0)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_username_input]").send_keys(" New")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_password_input]").send_keys("Password New")
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_save_button]").click()
        time.sleep(1)

        # Make sure the encryption key matches the one in the ui
        crypto_util.set_key(ui_util.ENCRYPTION_KEY)

        credential = db_util.get_credential("Credential Name")
        self.assertEqual(credential.username, "Username New")
        self.assertEqual(credential.password, "Password New")

    @ui_util.screenshot_on_error
    def test_delete_constant(self):
        """Test deleting a constant."""
        # Create a constant
        db_util.create_constant("Constant Name", "Value")
        ui_util.refresh_ui(self.browser)

        # Click constant in table
        ui_util.click_table_row(self.browser, "constants_tab_constants_table", 0)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=constant_popup_delete_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=popup_option1_button]").click()
        time.sleep(1)

        with self.assertRaises(ValueError):
            db_util.get_constant("Constant Name")

    @ui_util.screenshot_on_error
    def test_delete_credential(self):
        """Test deleting a credential."""
        # Create a credential
        db_util.create_credential("Credential Name", "Username", "Password")
        ui_util.refresh_ui(self.browser)

        # Click credential in table
        ui_util.click_table_row(self.browser, "constants_tab_credentials_table", 0)

        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=credential_popup_delete_button]").click()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=popup_option1_button]").click()
        time.sleep(1)

        with self.assertRaises(ValueError):
            db_util.get_credential("Credential Name")


if __name__ == '__main__':
    unittest.main()



"""This module is responsible for the layout and functionality of the 'New single Trigger' popup."""

from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime

from nicegui import ui
from cronsim import CronSim, CronSimError

from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import Trigger, TriggerStatus, TriggerType, ScheduledTrigger, SingleTrigger, QueueTrigger
from OpenOrchestrator.orchestrator.popups import generic_popups
from OpenOrchestrator.orchestrator import test_helper

if TYPE_CHECKING:
    from OpenOrchestrator.orchestrator.tabs.trigger_tab import TriggerTab


# pylint: disable-next=(too-many-instance-attributes, too-few-public-methods)
class TriggerPopup():
    """A popup for creating/updating triggers."""
    def __init__(self, trigger_tab: TriggerTab, trigger_type: TriggerType, trigger: Trigger | None = None):
        """Create a new popup.
        If a trigger is given it will be updated instead of creating a new trigger.

        Args:
            trigger_tab: The tab parent of the popup.
            trigger_type: The type of trigger popup to show.
            trigger: The Trigger to update if any.
        """
        self.trigger_tab = trigger_tab
        self.trigger_type = trigger_type
        self.trigger = trigger
        title = f'Update {trigger_type.value} Trigger' if trigger else f'New {trigger_type.value} Trigger'

        with ui.dialog(value=True) as self.dialog, ui.card().classes('w-full'):
            ui.label(title).classes("text-xl")
            self.trigger_input = ui.input("Trigger Name").classes("w-full")
            self.name_input = ui.input("Process Name").classes("w-full")
            self.cron_input = ui.input("Cron expression", on_change=self._cron_change).classes("w-full")  # For scheduled triggers
            self.time_input = DatetimeInput("Trigger Time")  # For scheduled/single triggers
            with self.cron_input:
                with ui.link(target="https://crontab.guru/", new_tab=True):
                    with ui.button(icon="help").props("flat dense"):
                        ui.tooltip("Help with cron: https://crontab.guru/")
            self.queue_input = ui.input("Queue Name").classes("w-full")  # For queue triggers
            self.batch_input = ui.number("Min Batch Size", value=1, min=1, precision=0, format="%.0f")  # For queue triggers
            self.path_input = ui.input("Process Path").classes("w-full")
            self.branch_input = ui.input("Git Branch/Tag (Optional)")
            self.git_check = ui.checkbox("Is path a Git Repo?")
            self.branch_input.bind_visibility_from(self.git_check, "value")
            self.args_input = ui.input("Process Arguments").classes("w-full")
            self.blocking_check = ui.checkbox(text="Is process blocking?", value=True)
            self.priority_input = ui.number("Priority", value=0, precision=0, format="%.0f")
            self.whitelist_input = ui.input_chips("Scheduler whitelist").classes("w-full")

            if trigger:
                with ui.row():
                    self.enable_button = ui.button("Enable", on_click=self._enable_trigger)
                    self.disable_button = ui.button("Disable", on_click=self._disable_trigger)
                    if trigger.process_status == TriggerStatus.RUNNING:
                        self.kill_button = ui.button("Kill", on_click=self._kill_trigger, color='red')
                    self.delete_button = ui.button("Delete", on_click=self._delete_trigger, color='red')
            else:
                # Dialog should only be persistent when a new trigger is being created
                self.dialog.props('persistent')

            with ui.row():
                self.save_button = ui.button("Save", on_click=self._create_trigger)
                self.cancel_button = ui.button("Cancel", on_click=self.dialog.close)

        self._disable_unused()
        self._define_validation()
        self._pre_populate()
        test_helper.set_automation_ids(self, "trigger_popup")

    def _define_validation(self):
        self.trigger_input._validation = {"Please enter a trigger name": bool}  # pylint: disable=protected-access
        self.name_input._validation = {"Please enter a process name": bool}  # pylint: disable=protected-access
        self.path_input._validation = {"Please enter a process path": bool}  # pylint: disable=protected-access
        self.queue_input._validation = {"Please enter a queue name": bool}  # pylint: disable=protected-access

        def validate_cron(value: str):
            try:
                CronSim(value, datetime.now())
                return True
            except CronSimError:
                return False

        self.cron_input._validation = {"Invalid cron expression": validate_cron}  # pylint: disable=protected-access

    def _pre_populate(self):
        """Populate the form with values from an existing trigger"""
        if not self.trigger:
            return

        self.trigger_input.value = self.trigger.trigger_name
        self.name_input.value = self.trigger.process_name
        self.path_input.value = self.trigger.process_path
        self.args_input.value = self.trigger.process_args
        self.git_check.value = self.trigger.is_git_repo
        self.branch_input.value = self.trigger.git_branch
        self.blocking_check.value = self.trigger.is_blocking
        self.priority_input.value = self.trigger.priority
        self.whitelist_input.value = self.trigger.scheduler_whitelist

        if isinstance(self.trigger, ScheduledTrigger):
            self.cron_input.value = self.trigger.cron_expr

        if isinstance(self.trigger, (SingleTrigger, ScheduledTrigger)):
            self.time_input.set_datetime(self.trigger.next_run)

        if isinstance(self.trigger, QueueTrigger):
            self.queue_input.value = self.trigger.queue_name
            self.batch_input.value = self.trigger.min_batch_size

    def _disable_unused(self):
        """Disable all inputs that aren't being used by the current trigger type."""
        if self.trigger_type == TriggerType.QUEUE:
            self.time_input.visible = False

        if self.trigger_type != TriggerType.SCHEDULED:
            self.cron_input.visible = False

        if self.trigger_type != TriggerType.QUEUE:
            self.queue_input.visible = False
            self.batch_input.visible = False

    def _cron_change(self):
        if self.cron_input.validate():
            cron_iter = CronSim(self.cron_input.value, datetime.now())
            self.time_input.set_datetime(next(cron_iter))

    async def _validate(self) -> bool:
        result = True

        result &= self.trigger_input.validate()
        result &= self.name_input.validate()
        result &= self.path_input.validate()

        if self.trigger_type in (TriggerType.SINGLE, TriggerType.SCHEDULED):
            result &= self.time_input.validate()

            next_run = self.time_input.get_datetime()
            if next_run and next_run < datetime.now():
                result &= await generic_popups.question_popup(
                        "The selected datetime is in the past. Do you want to create the trigger anyway?",
                        "Create", "Cancel")

        if self.trigger_type == TriggerType.SCHEDULED:
            result &= self.cron_input.validate()

        if self.trigger_type == TriggerType.QUEUE:
            result &= self.queue_input.validate()

        return result

    async def _create_trigger(self):
        """Creates a new single trigger in the database using the data entered in the UI.
        If an existing trigger was given when creating the popup it is updated instead.
        """
        if not await self._validate():
            ui.notify("Please fill out required information.", type='warning')
            return

        trigger_name = self.trigger_input.value
        process_name = self.name_input.value
        next_run: datetime = self.time_input.get_datetime()  # type: ignore
        cron_expr = self.cron_input.value
        queue_name = self.queue_input.value
        min_batch_size = self.batch_input.value
        path = self.path_input.value
        git_branch = self.branch_input.value
        args = self.args_input.value
        is_git = self.git_check.value
        is_blocking = self.blocking_check.value
        priority = self.priority_input.value
        whitelist = self.whitelist_input.value

        if self.trigger is None:
            # Create new trigger in database
            if self.trigger_type == TriggerType.SINGLE:
                db_util.create_single_trigger(trigger_name, process_name, next_run, path, args, is_git, is_blocking, priority, whitelist, git_branch)
            elif self.trigger_type == TriggerType.SCHEDULED:
                db_util.create_scheduled_trigger(trigger_name, process_name, cron_expr, next_run, path, args, is_git, is_blocking, priority, whitelist, git_branch)
            elif self.trigger_type == TriggerType.QUEUE:
                db_util.create_queue_trigger(trigger_name, process_name, queue_name, path, args, is_git, is_blocking, min_batch_size, priority, whitelist, git_branch)

            ui.notify("Trigger created", type='positive')
        else:
            # Update existing trigger
            self.trigger.trigger_name = trigger_name
            self.trigger.process_name = process_name
            self.trigger.process_path = path
            self.trigger.process_args = args
            self.trigger.is_git_repo = is_git
            self.trigger.is_blocking = is_blocking
            self.trigger.priority = priority
            self.trigger.scheduler_whitelist = whitelist
            self.trigger.git_branch = git_branch

            if isinstance(self.trigger, SingleTrigger):
                self.trigger.next_run = next_run
            elif isinstance(self.trigger, ScheduledTrigger):
                self.trigger.cron_expr = cron_expr
                self.trigger.next_run = next_run
            elif isinstance(self.trigger, QueueTrigger):
                self.trigger.queue_name = queue_name
                self.trigger.min_batch_size = min_batch_size

            db_util.update_trigger(self.trigger)
            ui.notify("Trigger updated", type='positive')

        self.dialog.close()
        self.trigger_tab.update()

    async def _delete_trigger(self):
        if not self.trigger:
            return

        if await generic_popups.question_popup(f"Delete trigger '{self.trigger.trigger_name}'?", "Delete", "Cancel", color1='red'):
            db_util.delete_trigger(self.trigger.id)
            ui.notify("Trigger deleted", type='positive')
            self.dialog.close()
            self.trigger_tab.update()

    def _disable_trigger(self):
        if not self.trigger:
            return

        if self.trigger.process_status == TriggerStatus.RUNNING:
            db_util.set_trigger_status(self.trigger.id, TriggerStatus.PAUSING)
        else:
            db_util.set_trigger_status(self.trigger.id, TriggerStatus.PAUSED)
        ui.notify("Trigger status set to 'Paused'.", type='positive')
        self.trigger_tab.update()

    def _kill_trigger(self):
        db_util.set_trigger_status(self.trigger.id, TriggerStatus.KILLING)
        ui.notify("Killing trigger", type='warning')
        self.trigger_tab.update()

    def _enable_trigger(self):
        if not self.trigger:
            return

        db_util.set_trigger_status(self.trigger.id, TriggerStatus.IDLE)
        ui.notify("Trigger status set to 'Idle'.", type='positive')
        self.trigger_tab.update()



# test_jobs_tab.py
"""Tests relating to the jobs tab in Orchestrator."""

import unittest
import time

from selenium.webdriver.common.by import By

from OpenOrchestrator.tests import db_test_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.jobs import JobStatus
from OpenOrchestrator.tests.ui_tests import ui_util


class TestJobsTab(unittest.TestCase):
    """Test functionality of the jobs tab ui."""

    def setUp(self) -> None:
        self.browser = ui_util.open_orchestrator()
        db_test_util.establish_clean_database()
        self.browser.find_element(By.CSS_SELECTOR, "[auto-id=jobs_tab]").click()
        ui_util.refresh_ui(self.browser)

    def tearDown(self) -> None:
        self.browser.quit()

    @ui_util.screenshot_on_error
    def test_jobs_table(self):
        """Test that jobs are shown correctly in the jobs table."""
        # Create some jobs
        self._create_test_jobs()

        ui_util.refresh_ui(self.browser)

        # Check result
        table_data = ui_util.get_table_data(self.browser, "jobs_tab_jobs_table")

        # Should have 3 jobs (newest first due to DESC ordering)
        self.assertEqual(len(table_data), 3)

        # Check first job (most recent - Failed)
        self.assertEqual(table_data[0][1], "Process3")
        self.assertEqual(table_data[0][2], "Scheduler1")
        self.assertRegex(table_data[0][3], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")  # Start time
        self.assertRegex(table_data[0][4], r"\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}")  # End time
        self.assertEqual(table_data[0][5], "Failed")

        # Check second job (Done)
        self.assertEqual(table_data[1][1], "Process2")
        self.assertEqual(table_data[1][5], "Done")

        # Check third job (Running - no end time)
        self.assertEqual(table_data[2][1], "Process1")
        self.assertEqual(table_data[2][4], "N/A")  # No end time for running job
        self.assertEqual(table_data[2][5], "Running")

    @ui_util.screenshot_on_error
    def test_job_click_navigates_to_logs(self):
        """Test that clicking a job row navigates to logs tab with job filter."""
        # Create a job
        job = db_util.start_job("Test Process", "Test Scheduler")
        db_util.set_job_status(job.id, JobStatus.DONE)

        ui_util.refresh_ui(self.browser)

        # Click the job row
        ui_util.click_table_row(self.browser, "jobs_tab_jobs_table", 0)

        # Should navigate to logs tab
        time.sleep(0.5)
        active_tab = self.browser.find_element(By.CSS_SELECTOR, ".q-tab.q-tab--active")
        self.assertEqual(active_tab.text, "LOGS")

        job_select = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=logs_tab_job_filter_label]")
        selected_value = job_select.text
        self.assertIn(str(job.id), selected_value)

    @ui_util.screenshot_on_error
    def test_empty_jobs_table(self):
        """Test that empty jobs table displays correctly."""
        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "jobs_tab_jobs_table")
        self.assertEqual(len(table_data), 0)

    @ui_util.screenshot_on_error
    def test_job_status_colors(self):
        """Test that job status badges have correct colors."""
        # Create jobs with different statuses
        db_util.start_job("Process1", "Scheduler")  # Running

        job2 = db_util.start_job("Process2", "Scheduler")
        db_util.set_job_status(job2.id, JobStatus.DONE)

        job3 = db_util.start_job("Process3", "Scheduler")
        db_util.set_job_status(job3.id, JobStatus.FAILED)

        job4 = db_util.start_job("Process4", "Scheduler")
        db_util.set_job_status(job4.id, JobStatus.KILLED)

        ui_util.refresh_ui(self.browser)

        # Get status badges
        table = self.browser.find_element(By.CSS_SELECTOR, "[auto-id=jobs_tab_jobs_table]")
        badges = table.find_elements(By.CSS_SELECTOR, ".q-badge")

        # Check we have 4 badges
        self.assertEqual(len(badges), 4)

        # Note: You might need to adjust color checking based on your implementation
        # This is a basic check that badges exist with text
        badge_texts = [b.text for b in badges]
        self.assertIn("Running", badge_texts)
        self.assertIn("Done", badge_texts)
        self.assertIn("Failed", badge_texts)
        self.assertIn("Killed", badge_texts)

    @ui_util.screenshot_on_error
    def test_jobs_sorted_by_start_time(self):
        """Test that jobs are sorted by start time descending."""
        # Create jobs with delays to ensure different timestamps
        db_util.start_job("First", "Scheduler")
        time.sleep(0.1)
        db_util.start_job("Second", "Scheduler")
        time.sleep(0.1)
        db_util.start_job("Third", "Scheduler")

        ui_util.refresh_ui(self.browser)

        table_data = ui_util.get_table_data(self.browser, "jobs_tab_jobs_table")

        # Newest should be first
        self.assertEqual(table_data[0][1], "Third")
        self.assertEqual(table_data[1][1], "Second")
        self.assertEqual(table_data[2][1], "First")

    def _create_test_jobs(self):
        """Create test jobs with different statuses."""
        # Running job
        db_util.start_job("Process1", "Scheduler1")
        time.sleep(0.1)

        # Done job
        job2 = db_util.start_job("Process2", "Scheduler2")
        db_util.set_job_status(job2.id, JobStatus.DONE)
        time.sleep(0.1)

        # Failed job
        job3 = db_util.start_job("Process3", "Scheduler1")
        db_util.set_job_status(job3.id, JobStatus.FAILED)

        ui_util.refresh_ui(self.browser)


if __name__ == '__main__':
    unittest.main()



"""This module is responsible for the layout and functionality of the 'New constant' popup."""

from __future__ import annotations
from typing import TYPE_CHECKING

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.constants import Credential
from OpenOrchestrator.orchestrator.popups.generic_popups import question_popup
from OpenOrchestrator.orchestrator import test_helper

if TYPE_CHECKING:
    from OpenOrchestrator.orchestrator.tabs.constants_tab import ConstantTab


# pylint: disable-next=too-few-public-methods, too-many-instance-attributes
class CredentialPopup():
    """A popup for creating/updating queue triggers."""
    def __init__(self, constant_tab: ConstantTab, credential: Credential | None = None):
        """Create a new popup.
        If a credential is given it will be updated instead of creating a new credential.

        Args:
            constant_tab: The tab that is opening this popup.
            credential: The credential to update if any.
        """
        self.constant_tab = constant_tab
        self.credential = credential
        title = 'Update Credential' if credential else 'New Credential'
        button_text = "Update" if credential else "Create"

        with ui.dialog(value=True).props('persistent') as self.dialog, ui.card().classes('w-full'):
            ui.label(title).classes("text-xl")
            self.name_input = ui.input("Credential Name").classes("w-full")
            self.username_input = ui.input("Username").classes("w-full")
            self.password_input = ui.input("Password").classes("w-full")

            with ui.row():
                self.save_button = ui.button(button_text, on_click=self._save_credential)
                self.cancel_button = ui.button("Cancel", on_click=self.dialog.close)

                if credential:
                    self.delete_button = ui.button("Delete", color='negative', on_click=self._delete_credential)

        self._define_validation()
        self._pre_populate()
        test_helper.set_automation_ids(self, "credential_popup")

    def _define_validation(self):
        """Define validation functions for ui elements."""
        self.name_input._validation = {"Please enter a name": bool}  # pylint: disable=protected-access
        self.username_input._validation = {"Please enter a username": bool}  # pylint: disable=protected-access
        self.password_input._validation = {"Please enter a password": bool}  # pylint: disable=protected-access

    def _pre_populate(self):
        """Pre populate the inputs with an existing credential."""
        if self.credential:
            self.name_input.value = self.credential.name
            self.name_input.disable()
            self.username_input.value = self.credential.username

    def _save_credential(self):
        """Create or update a credential in the database using the data from the UI."""
        self.name_input.validate()
        self.username_input.validate()
        self.password_input.validate()

        if self.name_input.error or self.username_input.error or self.password_input.error:
            return

        name = self.name_input.value
        username = self.username_input.value
        password = self.password_input.value

        if self.credential:
            db_util.update_credential(name, username, password)
        else:
            # Check if credential already exists
            try:
                db_util.get_credential(name, decrypt_password=False)
                exists = True
            except ValueError:
                exists = False

            if exists:
                ui.notify("A credential with that name already exists.", type='negative')
                return

            db_util.create_credential(name, username, password)

        self.dialog.close()
        self.constant_tab.update()

    async def _delete_credential(self):
        """Delete the selected credential."""
        if not self.credential:
            return
        if await question_popup(f"Delete credential '{self.credential.name}'?", "Delete", "Cancel", color1='negative'):
            db_util.delete_credential(self.credential.name)
            self.dialog.close()
            self.constant_tab.update()



"""This module is responsible for the layout and functionality of the Constants tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.orchestrator.popups.constant_popup import ConstantPopup
from OpenOrchestrator.orchestrator.popups.credential_popup import CredentialPopup
from OpenOrchestrator.orchestrator import test_helper

CONSTANT_COLUMNS = ("Constant Name", "Value", "Last Changed")
CREDENTIAL_COLUMNS = ("Credential Name", "Username", "Password", "Last Changed")


class ConstantTab():
    """The 'Constants' tab object."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            with ui.row():
                self.constant_button = ui.button("New Constant", icon='add', on_click=lambda e: ConstantPopup(self))
                self.credential_button = ui.button("New Credential", icon='add', on_click=lambda e: CredentialPopup(self))

            columns = [{'name': label, 'label': label, 'field': label, 'align': 'left', 'sortable': True} for label in CONSTANT_COLUMNS]
            self.constants_table = ui.table(title="Constants", columns=columns, rows=[], row_key='Constant Name', pagination=10).classes("w-full")
            self.constants_table.on('rowClick', self.row_click_constant)

            columns = [{'name': label, 'label': label, 'field': label, 'align': 'left', 'sortable': True} for label in CREDENTIAL_COLUMNS]
            self.credentials_table = ui.table(title="Credentials", columns=columns, rows=[], row_key='Credential Name', pagination=10).classes("w-full")
            self.credentials_table.on('rowClick', self.row_click_credential)

        test_helper.set_automation_ids(self, "constants_tab")

    def row_click_constant(self, event):
        """Callback for when a row is clicked in the table."""
        row = event.args[1]
        name = row['Constant Name']
        constant = db_util.get_constant(name)
        ConstantPopup(self, constant)

    def row_click_credential(self, event):
        """Callback for when a row is clicked in the table."""
        row = event.args[1]
        name = row['Credential Name']
        credential = db_util.get_credential(name, False)
        CredentialPopup(self, credential)

    def update(self):
        """Updates the tables on the tab."""
        constants = db_util.get_constants()
        self.constants_table.rows = [c.to_row_dict() for c in constants]
        self.constants_table.update()

        credentials = db_util.get_credentials()
        self.credentials_table.rows = [c.to_row_dict() for c in credentials]
        self.credentials_table.update()



"""This module is used to run Orchestrator or Scheduler from the command line."""

import argparse
import subprocess

from OpenOrchestrator.scheduler.application import Application as s_app
from OpenOrchestrator.orchestrator.application import Application as o_app


def main():
    """The main entry point of the CLI."""
    parser = argparse.ArgumentParser(
        prog="OpenOrchestrator",
        description="OpenOrchestrator is used to orchestrate, monitor and run Python-based automation scripts in Windows."
    )

    subparsers = parser.add_subparsers(title="Subcommands", required=True)

    o_parser = subparsers.add_parser("orchestrator", aliases=["o"], help="Start the Orchestrator application.")
    o_parser.add_argument("-p", "--port", type=int, help="Set the desired port for Orchestrator.")
    o_parser.add_argument("-d", "--dont_show", action="store_false", help="Set if you don't want Orchestrator to open in the browser automatically.")
    o_parser.set_defaults(func=orchestrator_command)

    s_parser = subparsers.add_parser("scheduler", aliases=["s"], help="Start the Scheduler application.")
    s_parser.set_defaults(func=scheduler_command)

    u_parser = subparsers.add_parser("upgrade", aliases=["u"], help="Upgrade the database to the newest revision or create a new database from scratch.")
    u_parser.add_argument("connection_string", type=str, help="The connection string to the database.")
    u_parser.add_argument("-n", "--new", action="store_true", help="Set if you're creating a new database from scratch.")
    u_parser.set_defaults(func=upgrade_command)

    args = parser.parse_args()
    args.func(args)


def orchestrator_command(args: argparse.Namespace):
    """Start the Orchestrator app.

    Args:
        args: The arguments Namespace object.
    """
    o_app(port=args.port, show=args.dont_show)


def scheduler_command(args: argparse.Namespace):  # pylint: disable=unused-argument
    """Start the Scheduler app."""
    s_app()


def upgrade_command(args: argparse.Namespace):
    """Upgrade the database to the newest revision using alembic.

    Args:
        args: The arguments Namespace object.
    """
    upgrade_args = ["alembic", "-x", args.connection_string, "upgrade", "head"]

    if args.new:
        subprocess.run(upgrade_args, check=True)
        print("Database upgraded to the newest revision!")
        return

    confirmation = input("Are you sure you want to upgrade the database to the newest revision? This cannot be undone. (y/n)").strip()

    if confirmation == "y":
        # Check if database is already stamped
        result = subprocess.run(["alembic", "-x", args.connection_string, "current"], check=True, capture_output=True)
        if result.stdout == b'':
            response = input("It seems like you are either creating a new database or trying to upgrade a legacy database. Please choose an option: (new/legacy/cancel)").strip()
            if response == "legacy":
                subprocess.run(["alembic", "-x", args.connection_string, "stamp", "b67b7649b282"], check=True)
            elif response != "new":
                print("Upgrade canceled")
                return

        subprocess.run(upgrade_args, check=True)
        print("Database upgraded to the newest revision!")
    else:
        print("Upgrade canceled")


if __name__ == '__main__':
    main()



"""Database revision '90d46abd44a3': Added trigger priority"""

from alembic import op
import sqlalchemy as sa


# pylint: disable=invalid-name
# revision identifiers, used by Alembic.
revision: str = '90d46abd44a3'
down_revision = '526b6edac328'
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade the database."""
    op.add_column('Triggers', sa.Column('scheduler_whitelist', sa.String(length=250), nullable=True))
    op.add_column('Triggers', sa.Column('priority', sa.Integer(), nullable=False, server_default=sa.text("0")))



"""This module defines ORM classes for constants and credentials."""

from datetime import datetime

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from OpenOrchestrator.common import datetime_util
from OpenOrchestrator.database.base import Base

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class Constant(Base):
    """A class representing a constant object in the ORM."""
    __tablename__ = "Constants"

    name: Mapped[str] = mapped_column(String(100), primary_key=True)
    value: Mapped[str] = mapped_column(String(1000))
    changed_at: Mapped[datetime] = mapped_column(onupdate=datetime.now, default=datetime.now)

    def to_row_dict(self) -> dict[str, str]:
        """Convert constant to a row dictionary for display in a table."""
        return {
            "Constant Name": self.name,
            "Value": self.value,
            "Last Changed": datetime_util.format_datetime(self.changed_at)
        }


class Credential(Base):
    """A class representing a credential object in the ORM."""
    __tablename__ = "Credentials"

    name: Mapped[str] = mapped_column(String(100), primary_key=True)
    username: Mapped[str] = mapped_column(String(250))
    password: Mapped[str] = mapped_column(String(1000))
    changed_at: Mapped[datetime] = mapped_column(onupdate=datetime.now, default=datetime.now)

    def to_row_dict(self) -> dict[str, str]:
        """Convert credential to a row dictionary for display in a table."""
        return {
            "Credential Name": self.name,
            "Username": self.username,
            "Password": self.format_password(),
            "Last Changed": datetime_util.format_datetime(self.changed_at)
        }

    def format_password(self) -> str:
        """Format the password to be shown in a table."""
        length = len(self.password)
        lower_length = int(((length-100)/20)*16)
        upper_length = lower_length + 15
        return f"{length} encrypted bytes. {lower_length}-{upper_length} decrypted bytes."
